"""
Framework integrations for Anchor.

1. Available integrations:
    - LangChain: AnchorMemory, AnchorChatHistory, AnchorCallbackHandler
    - CrewAI: AnchorCrewAgent, AnchorCrewMemory
    - Mem0: AnchorMem0

2. Usage:
    # LangChain
    from anchor.integrations.langchain import AnchorMemory, AnchorCallbackHandler

    # CrewAI
    from anchor.integrations.crewai import AnchorCrewAgent, AnchorCrewMemory

    # Mem0
    from anchor.integrations.mem0 import AnchorMem0

Note: Each integration requires its respective framework to be installed.
    pip install langchain-core  # for LangChain
    pip install crewai  # for CrewAI
    pip install mem0ai  # for Mem0
"""

__all__ = []


# Lazy imports to avoid requiring all framework dependencies
def __getattr__(name):
    if name == "langchain":
        from . import langchain

        return langchain
    elif name == "crewai":
        from . import crewai

        return crewai
    elif name == "mem0":
        from . import mem0

        return mem0
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



"""Data namespace: Policy-checked key-value storage with audit logging."""

from typing import Optional, Dict, Any, List
from dataclasses import dataclass, field
from datetime import datetime

from .base import BaseNamespace
from ..exceptions import NotFoundError


@dataclass
class WriteResult:
    """Result of a data write operation."""

    key: str
    allowed: bool
    audit_id: str
    blocked_by: Optional[str] = None
    reason: Optional[str] = None
    expires_at: Optional[datetime] = None
    created_at: Optional[datetime] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "WriteResult":
        return cls(
            key=data.get("key", ""),
            allowed=data.get("allowed", True),
            audit_id=data.get("audit_id", ""),
            blocked_by=data.get("blocked_by"),
            reason=data.get("reason"),
            expires_at=(
                datetime.fromisoformat(data["expires_at"].replace("Z", "+00:00"))
                if data.get("expires_at")
                else None
            ),
            created_at=(
                datetime.fromisoformat(data["created_at"].replace("Z", "+00:00"))
                if data.get("created_at")
                else None
            ),
        )


@dataclass
class DataEntry:
    """Full data entry with metadata."""

    key: str
    value: str
    metadata: Dict[str, Any] = field(default_factory=dict)
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    expires_at: Optional[datetime] = None
    audit_id: str = ""

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "DataEntry":
        return cls(
            key=data.get("key", ""),
            value=data.get("value", ""),
            metadata=data.get("metadata", {}),
            created_at=(
                datetime.fromisoformat(data["created_at"].replace("Z", "+00:00"))
                if data.get("created_at")
                else None
            ),
            updated_at=(
                datetime.fromisoformat(data["updated_at"].replace("Z", "+00:00"))
                if data.get("updated_at")
                else None
            ),
            expires_at=(
                datetime.fromisoformat(data["expires_at"].replace("Z", "+00:00"))
                if data.get("expires_at")
                else None
            ),
            audit_id=data.get("audit_id", ""),
        )


@dataclass
class SearchResult:
    """Search result with similarity score."""

    key: str
    value: str
    similarity: float
    metadata: Dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "SearchResult":
        return cls(
            key=data.get("key", ""),
            value=data.get("value", ""),
            similarity=data.get("similarity", 0.0),
            metadata=data.get("metadata", {}),
        )


class DataNamespace(BaseNamespace):
    """Key-value storage with policy checks and audit logging."""

    def write(
        self,
        agent_id: str,
        key: str,
        value: str,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> WriteResult:
        """
        Write a key-value pair. Policy-checked and audit-logged.

        Args:
            agent_id: Agent ID
            key: Key (e.g., "user:123:preference")
            value: Value to store
            metadata: Optional metadata

        Returns:
            WriteResult with allowed status and audit_id

        Note:
            If blocked by policy, result.allowed will be False
            and result.blocked_by will indicate which policy.
        """
        response = self._http.post(
            f"/agents/{agent_id}/data",
            data={
                "key": key,
                "value": value,
                "metadata": metadata or {},
            },
        )
        return WriteResult.from_dict(response)

    def write_batch(
        self,
        agent_id: str,
        items: Dict[str, str],
    ) -> List[WriteResult]:
        """
        Write multiple key-value pairs.

        Args:
            agent_id: Agent ID
            items: Dict of key -> value pairs

        Returns:
            List of WriteResult objects
        """
        response = self._http.post(
            f"/agents/{agent_id}/data/batch",
            data={"items": [{"key": k, "value": v} for k, v in items.items()]},
        )
        return [WriteResult.from_dict(r) for r in response.get("results", [])]

    def read(self, agent_id: str, key: str) -> Optional[str]:
        """
        Read a value by key. Audit-logged.

        Args:
            agent_id: Agent ID
            key: Key to read

        Returns:
            Value string, or None if not found
        """
        try:
            response = self._http.get(f"/agents/{agent_id}/data/{key}")
            return response.get("value")
        except NotFoundError:
            return None

    def read_full(self, agent_id: str, key: str) -> Optional[DataEntry]:
        """
        Read full entry including metadata.

        Args:
            agent_id: Agent ID
            key: Key to read

        Returns:
            DataEntry object, or None if not found
        """
        try:
            response = self._http.get(f"/agents/{agent_id}/data/{key}")
            return DataEntry.from_dict(response)
        except NotFoundError:
            return None

    def delete(self, agent_id: str, key: str) -> bool:
        """
        Delete a key. Audit-logged.

        Args:
            agent_id: Agent ID
            key: Key to delete

        Returns:
            True if deleted
        """
        self._http.delete(f"/agents/{agent_id}/data/{key}")
        return True

    def delete_prefix(self, agent_id: str, prefix: str) -> int:
        """
        Delete all keys with a prefix.

        Args:
            agent_id: Agent ID
            prefix: Key prefix (e.g., "user:123:")

        Returns:
            Number of keys deleted
        """
        response = self._http.delete(
            f"/agents/{agent_id}/data", params={"prefix": prefix}
        )
        return response.get("deleted_count", 0)

    def list(
        self,
        agent_id: str,
        prefix: Optional[str] = None,
        limit: int = 100,
    ) -> List[str]:
        """
        List keys.

        Args:
            agent_id: Agent ID
            prefix: Optional key prefix filter
            limit: Max keys to return

        Returns:
            List of key strings
        """
        params: Dict[str, Any] = {"limit": limit}
        if prefix:
            params["prefix"] = prefix
        response = self._http.get(f"/agents/{agent_id}/data", params=params)
        return [k.get("key", "") for k in response.get("keys", [])]

    def search(
        self,
        agent_id: str,
        query: str,
        limit: int = 10,
        prefix: Optional[str] = None,
        min_similarity: float = 0.7,
    ) -> List[SearchResult]:
        """
        Search data using vector similarity.

        Args:
            agent_id: Agent ID
            query: Search query
            limit: Max results
            prefix: Optional key prefix filter
            min_similarity: Minimum similarity threshold (0-1)

        Returns:
            List of SearchResult objects with similarity scores
        """
        data: Dict[str, Any] = {
            "query": query,
            "limit": limit,
            "min_similarity": min_similarity,
        }
        if prefix:
            data["prefix"] = prefix

        response = self._http.post(f"/agents/{agent_id}/data/search", data=data)
        return [SearchResult.from_dict(r) for r in response.get("results", [])]



"""Anchor: Control what your AI agents store. Audit everything.

Anchor lets you persist agent state, block bad data, prove what happened.

1. Quick Start:
    from anchor import Anchor

    anchor = Anchor(api_key="your-api-key")
    agent = anchor.agents.create("my-agent")

2. Full Example:
    from anchor import Anchor

    anchor = Anchor(api_key="your-api-key")

    # Create agent
    agent = anchor.agents.create(name="support-bot")

    # Store data (automatically audit-logged)
    anchor.data.write(agent.id, "user:123:pref", "concise answers")

    # PII gets blocked by policy
    result = anchor.data.write(agent.id, "user:123:ssn", "123-45-6789")
    print(result.allowed)  # False - blocked by policy

    # Checkpoint before risky operation
    checkpoint = anchor.checkpoints.create(agent.id, label="pre-update")

    # Something goes wrong? Rollback
    anchor.checkpoints.restore(agent.id, checkpoint.id)

    # Prove what happened
    events = anchor.audit.query(agent.id, limit=5)
"""

# Main client
from .anchor import Anchor
from .config import Config as ClientConfig

# Exceptions
from .exceptions import (
    AnchorError,
    AnchorAPIError,
    AuthenticationError,
    AuthorizationError,
    NotFoundError,
    ValidationError,
    PolicyViolationError,
    RateLimitError,
    ServerError,
    NetworkError,
    ChainIntegrityError,
)

# Types from namespaces
from .namespaces import (
    # Agent types
    Agent,
    # Config types
    Config,
    ConfigVersion,
    # Data types
    WriteResult,
    DataEntry,
    SearchResult,
    # Checkpoint types
    Checkpoint,
    RestoreResult,
    # Audit types
    AuditEvent,
    Verification,
    ExportResult,
)

__version__ = "1.0.0"

__all__ = [
    # Main client
    "Anchor",
    "ClientConfig",
    # Agent types
    "Agent",
    # Config types
    "Config",
    "ConfigVersion",
    # Data types
    "WriteResult",
    "DataEntry",
    "SearchResult",
    # Checkpoint types
    "Checkpoint",
    "RestoreResult",
    # Audit types
    "AuditEvent",
    "Verification",
    "ExportResult",
    # Exceptions
    "AnchorError",
    "AnchorAPIError",
    "AuthenticationError",
    "AuthorizationError",
    "NotFoundError",
    "ValidationError",
    "PolicyViolationError",
    "RateLimitError",
    "ServerError",
    "NetworkError",
    "ChainIntegrityError",
]



"""HTTP client utilities for Anchor SDK."""

import time
import requests
from typing import Optional, Dict, Any, TYPE_CHECKING

from .exceptions import (
    AnchorAPIError,
    AuthenticationError,
    AuthorizationError,
    NotFoundError,
    ValidationError,
    RateLimitError,
    ServerError,
    NetworkError,
    PolicyViolationError,
)

if TYPE_CHECKING:
    from .config import Config


class HttpClient:
    """
    HTTP client with retry logic and error handling.

    Used internally by all namespace classes.
    """

    def __init__(self, config: "Config"):
        self.config = config
        self.session = requests.Session()
        self.session.headers.update(
            {"X-API-Key": config.api_key, "User-Agent": "anchorai-python/1.0.0"}
        )

    def _handle_error(self, response: requests.Response) -> None:
        """Parse and raise appropriate exception for API errors"""
        status_code = response.status_code

        try:
            error_data = response.json()
            message = error_data.get("message") or error_data.get(
                "error", "Unknown error"
            )
        except (ValueError, KeyError):
            message = response.text or f"HTTP {status_code}"
            error_data = {}

        # Check for policy violation
        if error_data.get("blocked_by") or error_data.get("policy_name"):
            policy_name = error_data.get("blocked_by") or error_data.get(
                "policy_name", "unknown"
            )
            raise PolicyViolationError(message, policy_name, status_code, error_data)

        if status_code == 400:
            field = error_data.get("field")
            raise ValidationError(message, status_code, error_data, field)
        elif status_code == 401:
            raise AuthenticationError(message, status_code, error_data)
        elif status_code == 403:
            required = error_data.get("required_permission")
            raise AuthorizationError(message, status_code, error_data, required)
        elif status_code == 404:
            raise NotFoundError(message, status_code, error_data)
        elif status_code == 429:
            retry_after = error_data.get("retry_after")
            raise RateLimitError(message, status_code, error_data, retry_after)
        elif 500 <= status_code < 600:
            raise ServerError(message, status_code, error_data)
        else:
            raise AnchorAPIError(message, status_code, error_data)

    def request(
        self,
        method: str,
        endpoint: str,
        data: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """
        Make HTTP request to Anchor API with retry logic.

        Args:
            method: HTTP method (GET, POST, PUT, PATCH, DELETE)
            endpoint: API endpoint path (e.g., "/agents")
            data: Optional request body
            params: Optional query parameters

        Returns:
            Response JSON data

        Raises:
            AnchorAPIError: For API errors
            NetworkError: For network errors
        """
        url = f"{self.config.base_url}{endpoint}"
        last_exception = None

        for attempt in range(self.config.retry_attempts + 1):
            try:
                response = self.session.request(
                    method,
                    url,
                    json=data if data is not None else None,
                    params=params,
                    timeout=self.config.timeout,
                    verify=self.config.verify_ssl,
                )

                if response.ok:
                    # Handle empty responses
                    if not response.content:
                        return {}
                    return response.json()
                else:
                    # Don't retry on client errors (4xx), except 429
                    if (
                        400 <= response.status_code < 500
                        and response.status_code != 429
                    ):
                        self._handle_error(response)

                    # Retry on server errors (5xx) and rate limits (429)
                    if attempt < self.config.retry_attempts:
                        delay = self.config.retry_delay * (2**attempt)
                        time.sleep(delay)
                        continue
                    else:
                        self._handle_error(response)

            except requests.exceptions.Timeout:
                last_exception = NetworkError(
                    f"Request timeout after {self.config.timeout}s"
                )
                if attempt < self.config.retry_attempts:
                    time.sleep(self.config.retry_delay * (2**attempt))
                    continue

            except requests.exceptions.ConnectionError as e:
                last_exception = NetworkError(f"Connection error: {str(e)}")
                if attempt < self.config.retry_attempts:
                    time.sleep(self.config.retry_delay * (2**attempt))
                    continue

            except requests.exceptions.RequestException as e:
                last_exception = NetworkError(f"Request failed: {str(e)}")
                if attempt < self.config.retry_attempts:
                    time.sleep(self.config.retry_delay * (2**attempt))
                    continue

        if last_exception:
            raise last_exception
        raise NetworkError("Request failed after retries")

    def get(
        self, endpoint: str, params: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """Make GET request"""
        return self.request("GET", endpoint, params=params)

    def post(
        self,
        endpoint: str,
        data: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Make POST request"""
        return self.request("POST", endpoint, data=data, params=params)

    def put(
        self,
        endpoint: str,
        data: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Make PUT request"""
        return self.request("PUT", endpoint, data=data, params=params)

    def patch(
        self,
        endpoint: str,
        data: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Make PATCH request"""
        return self.request("PATCH", endpoint, data=data, params=params)

    def delete(
        self, endpoint: str, params: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """Make DELETE request"""
        return self.request("DELETE", endpoint, params=params)



"""Audit namespace for Anchor SDK (hash-chained audit trail per agent)."""

from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional, List, Dict, Any

from .base import BaseNamespace


@dataclass
class AuditEvent:
    """Audit event per API spec."""

    id: str
    agent_id: str
    operation: str
    resource: str
    result: str  # "allowed" | "blocked"
    blocked_by: Optional[str] = None
    timestamp: Optional[datetime] = None
    hash: str = ""
    previous_hash: Optional[str] = None
    metadata: Dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "AuditEvent":
        """Create AuditEvent from API response dict."""
        return cls(
            id=data.get("id", ""),
            agent_id=data.get("agent_id", ""),
            operation=data.get("operation", ""),
            resource=data.get("resource", ""),
            result=data.get("result", "allowed"),
            blocked_by=data.get("blocked_by"),
            timestamp=(
                datetime.fromisoformat(data["timestamp"].replace("Z", "+00:00"))
                if data.get("timestamp")
                else None
            ),
            hash=data.get("hash", ""),
            previous_hash=data.get("previous_hash"),
            metadata=data.get("metadata", {}),
        )


@dataclass
class Verification:
    """Result of audit chain verification."""

    valid: bool
    events_checked: int
    chain_start: Optional[str] = None
    chain_end: Optional[str] = None
    verified_at: Optional[datetime] = None
    first_invalid: Optional[Dict[str, Any]] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Verification":
        """Create Verification from API response dict."""
        return cls(
            valid=data.get("valid", False),
            events_checked=data.get("events_checked", 0),
            chain_start=data.get("chain_start"),
            chain_end=data.get("chain_end"),
            verified_at=(
                datetime.fromisoformat(data["verified_at"].replace("Z", "+00:00"))
                if data.get("verified_at")
                else None
            ),
            first_invalid=data.get("first_invalid"),
        )


@dataclass
class ExportResult:
    """Result of audit export request."""

    export_id: str
    format: str
    download_url: str
    expires_at: Optional[datetime] = None
    event_count: int = 0
    verification: Optional[Verification] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ExportResult":
        """Create ExportResult from API response dict."""
        verification = None
        if data.get("verification"):
            verification = Verification.from_dict(data["verification"])

        return cls(
            export_id=data.get("export_id", ""),
            format=data.get("format", "json"),
            download_url=data.get("download_url", ""),
            expires_at=(
                datetime.fromisoformat(data["expires_at"].replace("Z", "+00:00"))
                if data.get("expires_at")
                else None
            ),
            event_count=data.get("event_count", 0),
            verification=verification,
        )


class AuditNamespace(BaseNamespace):
    """
    Agent-scoped audit trail operations.

    All agent operations are logged to a hash-chained audit trail
    that can be verified for integrity.

    Usage:
        anchor = Anchor(api_key="your-api-key")

        # Query audit events for an agent
        events = anchor.audit.query(
            agent_id="agent_a1b2c3...",
            operations=["data.write", "data.delete"],
            limit=100
        )

        for event in events:
            print(f"{event.timestamp}: {event.operation}")
            print(f"  Result: {event.result}")
            print(f"  Hash: {event.hash}")

        # Verify chain integrity
        verification = anchor.audit.verify(agent_id="agent_a1b2c3...")
        print(verification.valid)  # True/False

        # Export for compliance
        export = anchor.audit.export(
            agent_id="agent_a1b2c3...",
            format="json",
            include_verification=True
        )
        print(f"Download: {export.download_url}")
    """

    def query(
        self,
        agent_id: str,
        operations: Optional[List[str]] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        limit: int = 100,
    ) -> List[AuditEvent]:
        """
        Query audit events for an agent.

        Args:
            agent_id: Agent ID (required)
            operations: Filter by operation types
                (e.g., ["data.write", "data.delete"])
            start_time: Filter events after this time
            end_time: Filter events before this time
            limit: Max events to return

        Returns:
            List of AuditEvent objects
        """
        params: Dict[str, Any] = {"limit": min(limit, 1000)}
        if operations:
            params["operations"] = ",".join(operations)
        if start_time:
            params["start"] = start_time.isoformat()
        if end_time:
            params["end"] = end_time.isoformat()

        response = self._http.get(f"/agents/{agent_id}/audit", params=params)
        return [AuditEvent.from_dict(e) for e in response.get("events", [])]

    def get(self, agent_id: str, audit_id: str) -> AuditEvent:
        """
        Get a specific audit event.

        Args:
            agent_id: Agent ID
            audit_id: Audit event ID

        Returns:
            AuditEvent object
        """
        response = self._http.get(f"/agents/{agent_id}/audit/{audit_id}")
        return AuditEvent.from_dict(response)

    def verify(
        self,
        agent_id: str,
        start_time: Optional[datetime] = None,
    ) -> Verification:
        """
        Verify hash chain integrity.

        Args:
            agent_id: Agent ID
            start_time: Optional start time for verification

        Returns:
            Verification result with valid status
        """
        params: Dict[str, Any] = {}
        if start_time:
            params["start"] = start_time.isoformat()

        response = self._http.get(f"/agents/{agent_id}/audit/verify", params=params)
        return Verification.from_dict(response)

    def export(
        self,
        agent_id: str,
        format: str = "json",
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        include_verification: bool = True,
    ) -> ExportResult:
        """
        Export audit trail for compliance.

        Args:
            agent_id: Agent ID
            format: Export format ("json" or "csv")
            start_time: Start of export range
            end_time: End of export range
            include_verification: Include chain verification

        Returns:
            ExportResult with download URL
        """
        payload: Dict[str, Any] = {
            "format": format,
            "include_verification": include_verification,
        }
        if start_time:
            payload["start"] = start_time.isoformat()
        if end_time:
            payload["end"] = end_time.isoformat()

        response = self._http.post(f"/agents/{agent_id}/audit/export", data=payload)
        return ExportResult.from_dict(response)



"""Anonymous telemetry for Anchor SDK

Tracks SDK usage to help improve the product. Completely opt-out, transparent, and privacy-respecting.
"""

import os
import platform
import sys
import threading
import time
from typing import Optional, Dict, Any
import requests

# Telemetry endpoint (can be overridden via env var)
TELEMETRY_ENDPOINT = os.getenv(
    "ANCHOR_TELEMETRY_ENDPOINT", "https://api.getanchor.dev/telemetry"
)

# Check if telemetry is disabled
TELEMETRY_ENABLED = os.getenv("ANCHOR_TELEMETRY", "1").lower() not in (
    "0",
    "false",
    "off",
    "no",
    "disabled",
)


class Telemetry:
    """Anonymous telemetry tracker for SDK usage"""

    def __init__(self, base_url: str, api_key: Optional[str] = None):
        self.base_url = base_url
        self.api_key = api_key
        self.enabled = TELEMETRY_ENABLED
        self.session = requests.Session()
        self.session.timeout = 2  # Short timeout, don't block

        # SDK metadata
        self.sdk_version = "1.0.0"
        self.language = "python"
        self.language_version = f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
        self.platform = platform.system()
        self.platform_version = platform.release()

        # Track initialization
        if self.enabled:
            self._track_async(
                "sdk.initialized",
                {
                    "sdk_version": self.sdk_version,
                    "language": self.language,
                    "language_version": self.language_version,
                    "platform": self.platform,
                    "platform_version": self.platform_version,
                    "has_api_key": bool(api_key),
                },
            )

    def _track_async(self, event: str, properties: Optional[Dict[str, Any]] = None):
        """Track an event asynchronously (non-blocking)"""
        if not self.enabled:
            return

        def send_event():
            try:
                payload = {
                    "event": event,
                    "properties": properties or {},
                    "sdk_version": self.sdk_version,
                    "language": self.language,
                    "timestamp": int(time.time() * 1000),
                }

                # Don't send API key or any sensitive data
                # Only send base_url domain (not full URL)
                if self.base_url:
                    try:
                        from urllib.parse import urlparse

                        parsed = urlparse(self.base_url)
                        payload["base_url_domain"] = parsed.netloc or "localhost"
                    except Exception:
                        pass

                self.session.post(
                    TELEMETRY_ENDPOINT,
                    json=payload,
                    timeout=2,
                    headers={"User-Agent": f"anchor-sdk-python/{self.sdk_version}"},
                )
            except Exception:
                # Silently fail - never interrupt user experience
                pass

        # Send in background thread
        thread = threading.Thread(target=send_event, daemon=True)
        thread.start()

    def track_method_call(
        self, method_name: str, success: bool = True, error_type: Optional[str] = None
    ):
        """Track a method call"""
        if not self.enabled:
            return

        properties = {"method": method_name, "success": success}

        if error_type:
            properties["error_type"] = error_type

        self._track_async("sdk.method_called", properties)

    def track_error(self, error_type: str, method_name: Optional[str] = None):
        """Track an error"""
        if not self.enabled:
            return

        properties = {"error_type": error_type}

        if method_name:
            properties["method"] = method_name

        self._track_async("sdk.error", properties)


# Global telemetry instance (created per client)
_telemetry_instances: Dict[str, Telemetry] = {}


def get_telemetry(base_url: str, api_key: Optional[str] = None) -> Telemetry:
    """Get or create telemetry instance for a client"""
    key = f"{base_url}:{api_key or 'no-key'}"
    if key not in _telemetry_instances:
        _telemetry_instances[key] = Telemetry(base_url, api_key)
    return _telemetry_instances[key]



"""
Anchor Mem0 Integration.

Adds policy enforcement, persistent storage, audit trails, and checkpoint/rollback to Mem0.

Anchor works ABOVE Mem0:
- Mem0 answers "what does the agent remember?"
- Anchor answers "what is the agent allowed to remember, and can you prove it?"

Usage:
    from anchor import Anchor
    from anchor.integrations.mem0 import AnchorMem0
    from mem0 import Memory

    anchor = Anchor(api_key="your-api-key")
    agent = anchor.agents.create("mem0-agent")

    # Configure policies
    anchor.config.update(agent.id, {
        "policies": {"block_pii": True, "retention_days": 90}
    })

    # Wrap Mem0 with policy enforcement
    mem0_client = Memory()
    wrapped = AnchorMem0(
        anchor=anchor,
        agent_id=agent.id,
        mem0_client=mem0_client
    )

    # Policy enforcement
    result = wrapped.add("User email is john@example.com", user_id="user_123")
    if not result.allowed:
        print(f"Blocked: {result.blocked_by}")

    # Checkpoint & rollback
    checkpoint_id = wrapped.create_checkpoint(label="before-batch")
    # wrapped.restore_checkpoint(checkpoint_id)  # if needed

    # Audit trail
    events = anchor.audit.query(agent.id, limit=10)
    verification = anchor.audit.verify(agent.id)  # Verify chain integrity
"""

from typing import Any, List, Optional, Dict
from dataclasses import dataclass

# Check for Mem0 availability
try:
    from mem0 import Memory

    MEM0_AVAILABLE = True
except ImportError:
    MEM0_AVAILABLE = False
    Memory = None


@dataclass
class PolicyResult:
    """Result of a policy-checked operation."""

    allowed: bool
    blocked_by: Optional[str] = None
    reason: Optional[str] = None
    mem0_result: Any = None
    expires_at: Optional[Any] = None  # datetime when data will auto-delete (if retention policy set)

    def __bool__(self) -> bool:
        return self.allowed


class AnchorMem0:
    """
    Policy-enforced Mem0 wrapper with persistent storage, audit trails, and checkpoints.

    Wraps a Mem0 Memory instance to add:
    - Policy checks before write operations
    - Automatic audit logging
    - PII/secrets blocking

    Usage:
        from anchor import Anchor
        from anchor.integrations.mem0 import AnchorMem0
        from mem0 import Memory

        anchor = Anchor(api_key="your-api-key")
        agent = anchor.agents.create(name="mem0-agent")

        # Configure policies
        anchor.config.update(agent.id, {
            "policies": {"block_pii": True, "block_secrets": True}
        })

        mem0_client = Memory()
        wrapped = AnchorMem0(
            anchor=anchor,
            agent_id=agent.id,
            mem0_client=mem0_client
        )

        # Operations are policy-checked
        result = wrapped.add("User prefers dark mode", user_id="user_123")
        if result.allowed:
            print("Stored successfully")
        else:
            print(f"Blocked by: {result.blocked_by}")
    """

    def __init__(self, anchor: Any, agent_id: str, mem0_client: Any):
        """
        Initialize AnchorMem0.

        Args:
            anchor: Anchor client instance
            agent_id: Anchor agent identifier
            mem0_client: Mem0 Memory instance
        """
        if not MEM0_AVAILABLE:
            raise ImportError("Mem0 is required. Install with: pip install mem0ai")

        self.anchor = anchor
        self.agent_id = agent_id
        self.mem0 = mem0_client

    def add(
        self,
        data: str,
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        run_id: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> PolicyResult:
        """
        Add memory with policy enforcement. Data persists across restarts.

        Args:
            data: Memory content
            user_id: User identifier
            agent_id: Mem0 agent identifier (not Anchor agent)
            run_id: Run identifier
            metadata: Additional metadata
            **kwargs: Additional Mem0 options

        Returns:
            PolicyResult with allowed status and expiration info (expires_at if retention set)
        """
        # Check policy by attempting to write to Anchor
        key = f"mem0:{user_id or 'unknown'}:{agent_id or 'default'}"
        write_result = self.anchor.data.write(
            self.agent_id,
            key,
            data,
            metadata={"source": "mem0", "user_id": user_id, "run_id": run_id},
        )

        if not write_result.allowed:
            return PolicyResult(
                allowed=False,
                blocked_by=write_result.blocked_by,
                reason=write_result.reason,
                mem0_result=None,
                expires_at=None,
            )

        # If allowed, call Mem0
        result = self.mem0.add(
            data=data,
            user_id=user_id,
            agent_id=agent_id,
            run_id=run_id,
            metadata=metadata,
            **kwargs,
        )

        # Return with expiration info from Anchor
        return PolicyResult(
            allowed=True, 
            blocked_by=None, 
            reason=None, 
            mem0_result=result,
            expires_at=write_result.expires_at,  # Include expiration date
        )

    def search(
        self,
        query: str,
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        run_id: Optional[str] = None,
        limit: int = 10,
        **kwargs,
    ) -> List[Dict[str, Any]]:
        """
        Search memories (not blocked, but audit-logged).

        Args:
            query: Search query
            user_id: User identifier
            agent_id: Mem0 agent identifier
            run_id: Run identifier
            limit: Maximum results
            **kwargs: Additional Mem0 options

        Returns:
            List of matching memories
        """
        # Log the search operation
        self.anchor.data.write(
            self.agent_id,
            f"mem0:search:{user_id or 'unknown'}",
            query[:200],
            metadata={"type": "search", "user_id": user_id},
        )

        return self.mem0.search(
            query=query,
            user_id=user_id,
            agent_id=agent_id,
            run_id=run_id,
            limit=limit,
            **kwargs,
        )

    def get(self, memory_id: str) -> Optional[Dict[str, Any]]:
        """
        Get a memory by ID.

        Args:
            memory_id: Memory identifier

        Returns:
            Memory or None if not found
        """
        return self.mem0.get(memory_id)

    def get_all(
        self,
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        run_id: Optional[str] = None,
        **kwargs,
    ) -> List[Dict[str, Any]]:
        """
        Get all memories.

        Args:
            user_id: Filter by user
            agent_id: Filter by Mem0 agent
            run_id: Filter by run
            **kwargs: Additional Mem0 options

        Returns:
            List of memories
        """
        return self.mem0.get_all(
            user_id=user_id, agent_id=agent_id, run_id=run_id, **kwargs
        )

    def update(self, memory_id: str, data: str, **kwargs) -> PolicyResult:
        """
        Update a memory with policy enforcement.

        Args:
            memory_id: Memory identifier
            data: New memory content
            **kwargs: Additional Mem0 options

        Returns:
            PolicyResult with allowed status and Mem0 result
        """
        # Check policy
        write_result = self.anchor.data.write(
            self.agent_id,
            f"mem0:update:{memory_id}",
            data,
            metadata={"type": "update", "memory_id": memory_id},
        )

        if not write_result.allowed:
            return PolicyResult(
                allowed=False,
                blocked_by=write_result.blocked_by,
                reason=write_result.reason,
                mem0_result=None,
                expires_at=None,
            )

        # If allowed, call Mem0
        result = self.mem0.update(memory_id=memory_id, data=data, **kwargs)

        return PolicyResult(
            allowed=True, 
            blocked_by=None, 
            reason=None, 
            mem0_result=result,
            expires_at=write_result.expires_at,  # Include expiration date
        )

    def delete(self, memory_id: str) -> Dict[str, Any]:
        """
        Delete a memory (audit-logged but not blocked).

        Args:
            memory_id: Memory identifier

        Returns:
            Deletion result
        """
        # Log deletion
        self.anchor.data.write(
            self.agent_id,
            f"mem0:delete:{memory_id}",
            "deleted",
            metadata={"type": "delete", "memory_id": memory_id},
        )

        return self.mem0.delete(memory_id)

    def delete_all(
        self,
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        run_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Delete all memories matching criteria.

        Args:
            user_id: Filter by user
            agent_id: Filter by Mem0 agent
            run_id: Filter by run

        Returns:
            Deletion result
        """
        # Log bulk deletion
        self.anchor.data.write(
            self.agent_id,
            f"mem0:delete_all:{user_id or 'all'}",
            "bulk_delete",
            metadata={"type": "delete_all", "user_id": user_id, "agent_id": agent_id},
        )

        return self.mem0.delete_all(user_id=user_id, agent_id=agent_id, run_id=run_id)

    def history(self, memory_id: str, **kwargs) -> List[Dict[str, Any]]:
        """
        Get memory history.

        Args:
            memory_id: Memory identifier
            **kwargs: Additional options

        Returns:
            List of memory history entries
        """
        return self.mem0.history(memory_id, **kwargs)

    def reset(self) -> Dict[str, Any]:
        """
        Reset all memories.

        Returns:
            Reset result
        """
        # Log reset
        self.anchor.data.write(
            self.agent_id, "mem0:reset", "full_reset", metadata={"type": "reset"}
        )

        return self.mem0.reset()

    def create_checkpoint(self, label: Optional[str] = None) -> str:
        """
        Create a checkpoint of current memory state.

        Args:
            label: Optional label for the checkpoint

        Returns:
            Checkpoint ID

        Example:
            # Create checkpoint before risky operation
            checkpoint_id = wrapped.create_checkpoint(label="before-migration")
            
            # ... perform risky operations ...
            
            # If something goes wrong, restore
            wrapped.restore_checkpoint(checkpoint_id)
        """
        checkpoint = self.anchor.checkpoints.create(
            self.agent_id, label=label or "mem0-checkpoint"
        )
        return checkpoint.id

    def restore_checkpoint(self, checkpoint_id: str) -> None:
        """
        Restore memory to a previous checkpoint.

        Args:
            checkpoint_id: ID of checkpoint to restore

        Example:
            # Restore to checkpoint after detecting corruption
            wrapped.restore_checkpoint(checkpoint_id)
            print("Memory restored to checkpoint")
        """
        self.anchor.checkpoints.restore(self.agent_id, checkpoint_id)
        # Note: Mem0's internal state may need to be reloaded after restore

    def query_audit_log(
        self,
        operations: Optional[List[str]] = None,
        limit: int = 100,
    ) -> List[Any]:
        """
        Query audit trail for this agent's operations.

        Args:
            operations: Filter by operation types (e.g., ["data.write"])
            limit: Max events to return

        Returns:
            List of audit events

        Example:
            # See what operations were performed
            events = wrapped.query_audit_log(operations=["data.write"], limit=50)
            for event in events:
                print(f"{event.timestamp}: {event.operation} - {event.result}")
        """
        return self.anchor.audit.query(
            self.agent_id, operations=operations, limit=limit
        )

    def verify_audit_chain(self) -> Any:
        """
        Verify the integrity of the audit chain.

        Returns:
            Verification result with valid status

        Example:
            # Verify logs haven't been tampered with
            verification = wrapped.verify_audit_chain()
            if verification.valid:
                print("Audit chain is intact")
            else:
                print("WARNING: Audit chain integrity compromised!")
        """
        return self.anchor.audit.verify(self.agent_id)

    def export_audit_log(
        self, format: str = "json", include_verification: bool = True
    ) -> Any:
        """
        Export audit trail for compliance.

        Args:
            format: Export format ("json" or "csv")
            include_verification: Include chain verification

        Returns:
            ExportResult with download URL

        Example:
            # Export for compliance review
            export = wrapped.export_audit_log(format="json")
            print(f"Download audit log: {export.download_url}")
        """
        return self.anchor.audit.export(
            self.agent_id, format=format, include_verification=include_verification
        )



"""
Anchor LangChain Integration.

Policy-checked memory, audit logging, and checkpoint/rollback for LangChain agents.

Components:
- AnchorMemory: LangChain memory with policy enforcement and checkpoints
- AnchorChatHistory: Drop-in replacement for chat history with policy checks
- AnchorCallbackHandler: Audit logging for chain operations

Usage:
    from anchor import Anchor
    from anchor.integrations.langchain import AnchorMemory, AnchorCallbackHandler
    from langchain.agents import AgentExecutor

    anchor = Anchor(api_key="your-api-key")
    agent = anchor.agents.create(name="langchain-agent")

    # Configure policies
    anchor.config.update(agent.id, {
        "policies": {"block_pii": True, "retention_days": 30}
    })

    # Create memory (uses anchor.data for storage)
    memory = AnchorMemory(anchor=anchor, agent_id=agent.id)

    # Create callback handler (logs to anchor.audit)
    callbacks = AnchorCallbackHandler(anchor=anchor, agent_id=agent.id)

    # Use with LangChain
    executor = AgentExecutor(agent=lc_agent, tools=tools, memory=memory)
    result =executor.invoke({"input": "Hello"}, config={"callbacks": [callbacks]})

    # Checkpoint & audit
    checkpoint_id = memory.create_checkpoint(label="before-experiment")
    events = anchor.audit.query(agent.id, limit=50)
    verification = anchor.audit.verify(agent.id)  # Verify chain integrity
"""

from typing import List, Optional, Any, Dict
from datetime import datetime, timezone

try:
    from langchain_core.chat_history import BaseChatMessageHistory
    from langchain_core.messages import (
        BaseMessage,
        HumanMessage,
        AIMessage,
        SystemMessage,
    )

    LANGCHAIN_AVAILABLE = True
except ImportError:
    LANGCHAIN_AVAILABLE = False
    BaseChatMessageHistory = object
    BaseMessage = object
    HumanMessage = object
    AIMessage = object
    SystemMessage = object


class AnchorMemory:
    """
    LangChain memory with policy enforcement, persistent storage, and checkpoints.

    Description: Stores conversation history using anchor.data with:
    - Policy enforcement (PII blocking, etc.)
    - Automatic audit logging
    - Checkpoint/rollback support

    Usage:
        from anchor import Anchor
        from anchor.integrations.langchain import AnchorMemory

        anchor = Anchor(api_key="your-api-key")
        agent = anchor.agents.create("my-agent")

        memory = AnchorMemory(anchor=anchor, agent_id=agent.id)

        # Use with LangChain
        chain = ConversationChain(llm=llm, memory=memory)
        checkpoint_id = memory.create_checkpoint(label="before-experiment")
    """

    def __init__(
        self,
        anchor: Any,
        agent_id: str,
        memory_key: str = "chat_history",
        return_messages: bool = True,
        session_id: Optional[str] = None,
    ):
        """
        Initialize AnchorMemory.

        Args:
            anchor: Anchor client instance
            agent_id: Agent identifier
            memory_key: Key to use in chain context
            return_messages: Return as messages (True) or string (False)
            session_id: Optional session ID for grouping messages
        """
        self.anchor = anchor
        self.agent_id = agent_id
        self.memory_key = memory_key
        self.return_messages = return_messages
        self.session_id = session_id or datetime.now(timezone.utc).strftime("%Y%m%d_%H%M%S")
        self._messages: List[Any] = []

    @property
    def memory_variables(self) -> List[str]:
        """Return memory variable names."""
        return [self.memory_key]

    def load_memory_variables(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Load memory for chain context."""
        # Load from Anchor data store
        self._load_from_anchor()

        if self.return_messages:
            return {self.memory_key: self._messages}
        else:
            return {self.memory_key: self._messages_to_string()}

    def save_context(self, inputs: Dict[str, Any], outputs: Dict[str, Any]) -> None:
        """Save conversation turn to Anchor (persistent storage, survives restarts)."""
        input_str = inputs.get("input", "")
        output_str = outputs.get("output", "")

        timestamp = datetime.now(timezone.utc).isoformat()

        if input_str:
            if LANGCHAIN_AVAILABLE:
                self._messages.append(HumanMessage(content=input_str))
            # Store in Anchor (policy-checked)
            self.anchor.data.write(
                self.agent_id,
                f"chat:{self.session_id}:human:{timestamp}",
                input_str,
                metadata={"role": "human", "session_id": self.session_id},
            )

        if output_str:
            if LANGCHAIN_AVAILABLE:
                self._messages.append(AIMessage(content=output_str))
            # Store in Anchor (policy-checked)
            self.anchor.data.write(
                self.agent_id,
                f"chat:{self.session_id}:ai:{timestamp}",
                output_str,
                metadata={"role": "ai", "session_id": self.session_id},
            )

    def clear(self) -> None:
        """Clear conversation history."""
        self._messages = []
        # Delete from Anchor
        self.anchor.data.delete_prefix(self.agent_id, f"chat:{self.session_id}:")

    def _load_from_anchor(self) -> None:
        """Load messages from Anchor (includes historical data from previous sessions)."""
        if not LANGCHAIN_AVAILABLE:
            return

        # List keys for this session
        keys = self.anchor.data.list(self.agent_id, prefix=f"chat:{self.session_id}:")

        self._messages = []
        for key in sorted(keys):
            entry = self.anchor.data.read_full(self.agent_id, key)
            if entry:
                role = (
                    entry.metadata.get("role", "human") if entry.metadata else "human"
                )
                if role == "human":
                    self._messages.append(HumanMessage(content=entry.value))
                elif role == "ai":
                    self._messages.append(AIMessage(content=entry.value))
                elif role == "system":
                    self._messages.append(SystemMessage(content=entry.value))

    def _messages_to_string(self) -> str:
        """Convert messages to string format."""
        lines = []
        for msg in self._messages:
            if LANGCHAIN_AVAILABLE:
                if isinstance(msg, HumanMessage):
                    lines.append(f"Human: {msg.content}")
                elif isinstance(msg, AIMessage):
                    lines.append(f"AI: {msg.content}")
                elif isinstance(msg, SystemMessage):
                    lines.append(f"System: {msg.content}")
        return "\n".join(lines)

    def create_checkpoint(self, label: Optional[str] = None) -> str:
        """
        Create a checkpoint of current conversation state.

        Args:
            label: Optional label for the checkpoint

        Returns:
            Checkpoint ID

        Example:
            # Create checkpoint before risky operation
            checkpoint_id = memory.create_checkpoint(label="before-migration")
            
            # ... perform risky operations ...
            
            # If corruption detected, restore
            memory.restore_checkpoint(checkpoint_id)
        """
        checkpoint = self.anchor.checkpoints.create(
            self.agent_id, label=label or f"langchain-{self.session_id}"
        )
        return checkpoint.id

    def restore_checkpoint(self, checkpoint_id: str) -> None:
        """
        Restore conversation to a previous checkpoint.

        Args:
            checkpoint_id: ID of checkpoint to restore

        Example:
            # Restore to checkpoint after detecting corruption
            memory.restore_checkpoint(checkpoint_id)
            print("Conversation restored to checkpoint")
        """
        self.anchor.checkpoints.restore(self.agent_id, checkpoint_id)
        self._messages = []
        self._load_from_anchor()


class AnchorChatHistory(BaseChatMessageHistory if LANGCHAIN_AVAILABLE else object):
    """
    LangChain ChatMessageHistory backed by Anchor.

    Description: Drop-in replacement for LangChain's chat history that stores
    messages in Anchor's policy-checked data store.

    Usage:
        from anchor import Anchor
        from anchor.integrations.langchain import AnchorChatHistory

        anchor = Anchor(api_key="your-api-key")
        agent = anchor.agents.create("chat-agent")

        history = AnchorChatHistory(anchor=anchor, agent_id=agent.id)

        # Use with RunnableWithMessageHistory
        with_history = RunnableWithMessageHistory(
            chain,
            lambda session_id: history,
            input_messages_key="input",
            history_messages_key="history"
        )
    """

    def __init__(self, anchor: Any, agent_id: str, session_id: Optional[str] = None):
        """
        Initialize AnchorChatHistory.

        Args:
            anchor: Anchor client instance
            agent_id: Agent identifier
            session_id: Optional session ID
        """
        if not LANGCHAIN_AVAILABLE:
            raise ImportError(
                "LangChain is required. Install with: pip install langchain-core"
            )

        self.anchor = anchor
        self.agent_id = agent_id
        self.session_id = session_id or datetime.now(timezone.utc).strftime("%Y%m%d_%H%M%S")

    @property
    def messages(self) -> List[BaseMessage]:
        """Retrieve all messages."""
        keys = self.anchor.data.list(self.agent_id, prefix=f"chat:{self.session_id}:")

        messages = []
        for key in sorted(keys):
            entry = self.anchor.data.read_full(self.agent_id, key)
            if entry:
                role = (
                    entry.metadata.get("role", "human") if entry.metadata else "human"
                )
                if role == "human":
                    messages.append(HumanMessage(content=entry.value))
                elif role == "ai":
                    messages.append(AIMessage(content=entry.value))
                elif role == "system":
                    messages.append(SystemMessage(content=entry.value))

        return messages

    def add_message(self, message: BaseMessage) -> None:
        """Add a message to history."""
        timestamp = datetime.now(timezone.utc).isoformat()

        if isinstance(message, HumanMessage):
            role = "human"
        elif isinstance(message, AIMessage):
            role = "ai"
        elif isinstance(message, SystemMessage):
            role = "system"
        else:
            role = "human"

        content = (
            message.content
            if isinstance(message.content, str)
            else str(message.content)
        )

        self.anchor.data.write(
            self.agent_id,
            f"chat:{self.session_id}:{role}:{timestamp}",
            content,
            metadata={"role": role, "session_id": self.session_id},
        )

    def clear(self) -> None:
        """Clear all messages."""
        self.anchor.data.delete_prefix(self.agent_id, f"chat:{self.session_id}:")


class AnchorCallbackHandler:
    """
    LangChain callback handler for audit logging.

    Description: Logs chain operations to Anchor's audit trail (hash-chained, verifiable).

    Usage:
        from anchor import Anchor
        from anchor.integrations.langchain import AnchorCallbackHandler

        anchor = Anchor(api_key="your-api-key")
        handler = AnchorCallbackHandler(anchor=anchor, agent_id="agent_123")

        # Use with chain
        result = chain.invoke(input, config={"callbacks": [handler]})

        # Query audit logs
        events = anchor.audit.query(agent_id="agent_123")
        verification = anchor.audit.verify(agent_id="agent_123")
    """

    def __init__(
        self,
        anchor: Any,
        agent_id: str,
        log_prompts: bool = True,
        log_completions: bool = True,
        log_tool_calls: bool = True,
    ):
        """
        Initialize callback handler.

        Args:
            anchor: Anchor client instance
            agent_id: Agent identifier
            log_prompts: Log input prompts
            log_completions: Log completions
            log_tool_calls: Log tool invocations
        """
        self.anchor = anchor
        self.agent_id = agent_id
        self.log_prompts = log_prompts
        self.log_completions = log_completions
        self.log_tool_calls = log_tool_calls

    def on_llm_start(
        self, serialized: Dict[str, Any], prompts: List[str], **kwargs
    ) -> None:
        """Called when LLM starts. Logs to audit trail (automatically hash-chained)."""
        if self.log_prompts:
            # Log to audit (operations are automatically logged to hash-chained audit trail)
            for i, prompt in enumerate(prompts):
                self.anchor.data.write(
                    self.agent_id,
                    f"llm:prompt:{datetime.now(timezone.utc).isoformat()}:{i}",
                    prompt[:1000],  # Truncate for storage
                    metadata={"type": "llm_prompt"},
                )

    def on_llm_end(self, response: Any, **kwargs) -> None:
        """Called when LLM completes. Logs to audit trail (automatically hash-chained)."""
        if self.log_completions:
            # Log completion (automatically added to hash-chained audit trail)
            output = str(response)[:1000] if response else ""
            self.anchor.data.write(
                self.agent_id,
                f"llm:completion:{datetime.now(timezone.utc).isoformat()}",
                output,
                metadata={"type": "llm_completion"},
            )

    def on_tool_start(
        self, serialized: Dict[str, Any], input_str: str, **kwargs
    ) -> None:
        """Called when tool starts. Logs to audit trail (automatically hash-chained)."""
        if self.log_tool_calls:
            tool_name = serialized.get("name", "unknown")
            self.anchor.data.write(
                self.agent_id,
                f"tool:start:{tool_name}:{datetime.now(timezone.utc).isoformat()}",
                input_str[:500],
                metadata={"type": "tool_start", "tool": tool_name},
            )

    def on_tool_end(self, output: str, **kwargs) -> None:
        """Called when tool completes. Logs to audit trail (automatically hash-chained)."""
        if self.log_tool_calls:
            self.anchor.data.write(
                self.agent_id,
                f"tool:end:{datetime.now(timezone.utc).isoformat()}",
                output[:500] if output else "",
                metadata={"type": "tool_end"},
            )



"""Agents namespace for Anchor SDK: agent registry and lifecycle management."""

from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional, List, Dict, Any

from .base import BaseNamespace


@dataclass
class Agent:
    """Agent record."""

    id: str
    name: str
    status: str = "active"  # "active" | "suspended"
    metadata: Dict[str, Any] = field(default_factory=dict)
    config_version: Optional[str] = None
    data_count: Optional[int] = None
    checkpoint_count: Optional[int] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Agent":
        """Create Agent from API response dict."""
        return cls(
            id=data.get("id") or data.get("agent_id", ""),
            name=data.get("name", ""),
            status=data.get("status", "active"),
            metadata=data.get("metadata") or data.get("config") or {},
            config_version=data.get("config_version"),
            data_count=data.get("data_count"),
            checkpoint_count=data.get("checkpoint_count"),
            created_at=(
                datetime.fromisoformat(data["created_at"].replace("Z", "+00:00"))
                if data.get("created_at")
                else None
            ),
            updated_at=(
                datetime.fromisoformat(data["updated_at"].replace("Z", "+00:00"))
                if data.get("updated_at")
                else None
            ),
        )


class AgentsNamespace(BaseNamespace):
    """
    Agent management operations.

    Usage:
        anchor = Anchor(api_key="your-api-key")

        # Create agent
        agent = anchor.agents.create(
            name="support-bot",
            metadata={"environment": "production", "version": "1.0.0"}
        )

        # Lifecycle operations
        agents = anchor.agents.list(status="active")
        agent = anchor.agents.get("agent_a1b2c3...")
        anchor.agents.suspend("agent_a1b2c3...")
        anchor.agents.activate("agent_a1b2c3...")
        anchor.agents.delete("agent_a1b2c3...")
    """

    def create(
        self,
        name: str,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> Agent:
        """
        Create a new agent.

        Args:
            name: Human-readable agent name
            metadata: Optional key-value metadata

        Returns:
            Created Agent object
        """
        response = self._http.post(
            "/agents",
            data={
                "name": name,
                "metadata": metadata or {},
            },
        )
        agent_data = response.get("agent", response)
        return Agent.from_dict(agent_data)

    def get(self, agent_id: str) -> Agent:
        """
        Get an agent by ID.

        Args:
            agent_id: Agent ID (e.g., "agent_a1b2c3")

        Returns:
            Agent object

        Raises:
            NotFoundError: If agent doesn't exist
        """
        response = self._http.get(f"/agents/{agent_id}")
        agent_data = response.get("agent", response)
        return Agent.from_dict(agent_data)

    def list(
        self,
        status: Optional[str] = None,
        limit: int = 50,
        offset: int = 0,
    ) -> List[Agent]:
        """
        List agents.

        Args:
            status: Filter by status ("active", "suspended")
            limit: Max results (default 50)
            offset: Pagination offset

        Returns:
            List of Agent objects
        """
        params: Dict[str, Any] = {"limit": min(limit, 100), "offset": offset}
        if status:
            params["status"] = status

        response = self._http.get("/agents", params=params)
        agents_data = response.get("agents", [])
        return [Agent.from_dict(a) for a in agents_data]

    def update(
        self,
        agent_id: str,
        metadata: Dict[str, Any],
    ) -> Agent:
        """
        Update agent metadata.

        Args:
            agent_id: Agent ID
            metadata: New metadata (merged with existing)

        Returns:
            Updated Agent object
        """
        response = self._http.patch(
            f"/agents/{agent_id}",
            data={
                "metadata": metadata,
            },
        )
        agent_data = response.get("agent", response)
        return Agent.from_dict(agent_data)

    def delete(self, agent_id: str) -> bool:
        """
        Delete an agent.

        Args:
            agent_id: Agent ID

        Returns:
            True if deleted
        """
        self._http.delete(f"/agents/{agent_id}")
        return True

    def suspend(self, agent_id: str) -> Agent:
        """
        Suspend an agent. Suspended agents cannot perform operations.

        Args:
            agent_id: Agent ID

        Returns:
            Updated Agent object
        """
        response = self._http.post(f"/agents/{agent_id}/suspend")
        agent_data = response.get("agent", response)
        return Agent.from_dict(agent_data)

    def activate(self, agent_id: str) -> Agent:
        """
        Activate a suspended agent.

        Args:
            agent_id: Agent ID

        Returns:
            Updated Agent object
        """
        response = self._http.post(f"/agents/{agent_id}/activate")
        agent_data = response.get("agent", response)
        return Agent.from_dict(agent_data)



"""Custom exceptions for Anchor SDK."""

from typing import Dict, Any


class AnchorError(Exception):
    """Base exception for Anchor SDK."""

    def __init__(self, message: str, code: str = None):
        super().__init__(message)
        self.message = message
        self.code = code


class AnchorAPIError(AnchorError):
    """Base exception for API errors."""

    def __init__(
        self, message: str, status_code: int = None, response: Dict[str, Any] = None
    ):
        super().__init__(message, code="api_error")
        self.status_code = status_code
        self.response = response or {}


class AuthenticationError(AnchorAPIError):
    """Raised when authentication fails (401)."""

    def __init__(
        self,
        message: str = "Invalid API key",
        status_code: int = 401,
        response: Dict[str, Any] = None,
    ):
        super().__init__(message, status_code, response)
        self.code = "authentication_error"


class AuthorizationError(AnchorAPIError):
    """Raised when authorization fails (403)."""

    def __init__(
        self,
        message: str,
        status_code: int = 403,
        response: Dict[str, Any] = None,
        required_permission: str = None,
    ):
        super().__init__(message, status_code, response)
        self.code = "authorization_error"
        self.required_permission = required_permission


class NotFoundError(AnchorAPIError):
    """Raised when a resource is not found (404)."""

    def __init__(
        self, message: str, status_code: int = 404, response: Dict[str, Any] = None
    ):
        super().__init__(message, status_code, response)
        self.code = "not_found"


class ValidationError(AnchorAPIError):
    """Raised when input validation fails (400)."""

    def __init__(
        self,
        message: str,
        status_code: int = 400,
        response: Dict[str, Any] = None,
        field: str = None,
    ):
        super().__init__(message, status_code, response)
        self.code = "validation_error"
        self.field = field


class PolicyViolationError(AnchorAPIError):
    """Raised when an operation is blocked by policy."""

    def __init__(
        self,
        message: str,
        policy_name: str,
        status_code: int = 403,
        response: Dict[str, Any] = None,
    ):
        super().__init__(message, status_code, response)
        self.code = "policy_violation"
        self.policy_name = policy_name


class RateLimitError(AnchorAPIError):
    """Raised when rate limited (429)."""

    def __init__(
        self,
        message: str,
        status_code: int = 429,
        response: Dict[str, Any] = None,
        retry_after: int = None,
    ):
        super().__init__(message, status_code, response)
        self.code = "rate_limit"
        self.retry_after = retry_after


class ServerError(AnchorAPIError):
    """Raised for server errors (5xx)."""

    def __init__(
        self, message: str, status_code: int = 500, response: Dict[str, Any] = None
    ):
        super().__init__(message, status_code, response)
        self.code = "server_error"


class NetworkError(AnchorError):
    """Raised for network connectivity issues."""

    def __init__(self, message: str):
        super().__init__(message, code="network_error")


class ChainIntegrityError(AnchorError):
    """Raised when audit chain integrity verification fails."""

    def __init__(self, message: str, entry_id: str = None):
        super().__init__(message, code="chain_integrity_error")
        self.entry_id = entry_id



"""Config namespace for agent configuration management."""

from typing import Optional, Dict, Any, List
from dataclasses import dataclass
from datetime import datetime

from .base import BaseNamespace


@dataclass
class Config:
    """Agent configuration with versioning."""

    agent_id: str
    version: str
    config: Dict[str, Any]
    previous_version: Optional[str] = None
    created_at: Optional[datetime] = None
    created_by: Optional[str] = None
    audit_id: Optional[str] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Config":
        return cls(
            agent_id=data.get("agent_id", ""),
            version=data.get("version", ""),
            config=data.get("config", {}),
            previous_version=data.get("previous_version"),
            created_at=(
                datetime.fromisoformat(data["created_at"].replace("Z", "+00:00"))
                if data.get("created_at")
                else None
            ),
            created_by=data.get("created_by"),
            audit_id=data.get("audit_id"),
        )


@dataclass
class ConfigVersion:
    """Config version summary."""

    version: str
    created_at: Optional[datetime] = None
    created_by: Optional[str] = None
    summary: Optional[str] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ConfigVersion":
        return cls(
            version=data.get("version", ""),
            created_at=(
                datetime.fromisoformat(data["created_at"].replace("Z", "+00:00"))
                if data.get("created_at")
                else None
            ),
            created_by=data.get("created_by"),
            summary=data.get("summary"),
        )


class ConfigNamespace(BaseNamespace):
    """Agent configuration management with versioning."""

    def get(self, agent_id: str) -> Config:
        """
        Get current config for an agent.

        Args:
            agent_id: Agent ID

        Returns:
            Current Config object
        """
        response = self._http.get(f"/agents/{agent_id}/config")
        return Config.from_dict(response)

    def update(self, agent_id: str, config: Dict[str, Any]) -> Config:
        """
        Update agent config. Creates a new version.

        Args:
            agent_id: Agent ID
            config: Config object with behavior, persistence, policies

        Returns:
            New Config object with version

        Example:
            anchor.config.update(agent_id, {
                "behavior": {
                    "instructions": "You are helpful",
                    "constraints": ["Be professional"]
                },
                "policies": {
                    "block_pii": True,
                    "retention_days": 90
                }
            })
        """
        response = self._http.put(f"/agents/{agent_id}/config", data=config)
        return Config.from_dict(response)

    def versions(self, agent_id: str, limit: int = 20) -> List[ConfigVersion]:
        """
        List config versions.

        Args:
            agent_id: Agent ID
            limit: Max versions to return

        Returns:
            List of ConfigVersion objects
        """
        response = self._http.get(
            f"/agents/{agent_id}/config/versions", params={"limit": limit}
        )
        return [ConfigVersion.from_dict(v) for v in response.get("versions", [])]

    def get_version(self, agent_id: str, version: str) -> Config:
        """
        Get a specific config version.

        Args:
            agent_id: Agent ID
            version: Version string (e.g., "v3")

        Returns:
            Config object for that version
        """
        response = self._http.get(f"/agents/{agent_id}/config/versions/{version}")
        return Config.from_dict(response)

    def rollback(self, agent_id: str, target_version: str) -> Config:
        """
        Rollback config to a previous version.
        Creates a new version with the old config.

        Args:
            agent_id: Agent ID
            target_version: Version to rollback to

        Returns:
            New Config object (copy of target version)
        """
        response = self._http.post(
            f"/agents/{agent_id}/config/rollback",
            data={"target_version": target_version},
        )
        return Config.from_dict(response)



"""Data models returned by Anchor SDK methods."""

from datetime import datetime
from typing import Any, Dict, Optional

from pydantic import BaseModel


class Agent(BaseModel):
    """Agent model."""

    id: str
    name: str
    status: str  # "active" | "suspended"
    metadata: Dict[str, Any] = {}
    config_version: Optional[str] = None
    data_count: Optional[int] = None
    checkpoint_count: Optional[int] = None
    created_at: datetime
    updated_at: datetime


class Config(BaseModel):
    """Agent configuration model."""

    agent_id: str
    version: str
    config: Dict[str, Any]  # e.g., { behavior: {...}, policies: {...} }
    previous_version: Optional[str] = None
    created_at: datetime
    created_by: Optional[str] = None
    audit_id: Optional[str] = None


class ConfigVersion(BaseModel):
    """Config version summary."""

    version: str
    created_at: datetime
    created_by: Optional[str] = None
    summary: Optional[str] = None


class WriteResult(BaseModel):
    """Result of a data write operation."""

    key: str
    allowed: bool
    audit_id: str
    blocked_by: Optional[str] = None  # e.g., "policy:block_pii"
    reason: Optional[str] = None
    expires_at: Optional[datetime] = None
    created_at: Optional[datetime] = None


class DataEntry(BaseModel):
    """Full data entry with metadata."""

    key: str
    value: str
    metadata: Dict[str, Any] = {}
    created_at: datetime
    updated_at: datetime
    expires_at: Optional[datetime] = None
    audit_id: str


class SearchResult(BaseModel):
    """Search result with similarity score."""

    key: str
    value: str
    similarity: float  # 0.0 - 1.0
    metadata: Dict[str, Any] = {}


class DataSnapshot(BaseModel):
    """Data snapshot info in checkpoint."""

    key_count: int
    total_size_bytes: int


class Checkpoint(BaseModel):
    """Checkpoint model."""

    id: str
    agent_id: str
    label: Optional[str] = None
    description: Optional[str] = None
    config_version: str
    data_snapshot: DataSnapshot
    created_at: datetime
    audit_id: Optional[str] = None


class RestoreResult(BaseModel):
    """Result of checkpoint restore."""

    restored_from: str
    config_restored: bool
    config_version: Optional[str] = None
    data_restored: bool
    data_keys_restored: int
    data_keys_removed: int
    audit_id: str
    restored_at: datetime


class AuditEvent(BaseModel):
    """Audit event model."""

    id: str
    agent_id: str
    operation: str  # e.g., # "data.write", "data.delete", "config.update", "checkpoint.create", "checkpoint.restore"
    resource: str  # e.g., "user:123:preference"
    result: str  # "allowed" | "blocked"
    blocked_by: Optional[str] = None
    timestamp: datetime
    hash: str  # sha256 hash
    previous_hash: Optional[str] = None
    metadata: Dict[str, Any] = {}


class Verification(BaseModel):
    """Audit chain verification result."""

    valid: bool
    events_checked: int
    chain_start: Optional[str] = None
    chain_end: Optional[str] = None
    verified_at: datetime
    first_invalid: Optional[Dict[str, Any]] = None


class ExportResult(BaseModel):
    """Audit export result."""

    export_id: str
    format: str
    download_url: str
    expires_at: datetime
    event_count: int
    verification: Optional[Verification] = None



"""
Anchor CrewAI Integration.

Policy-checked agents, storage, audit trails, and checkpoints for CrewAI multi-agent systems.

Components:
- AnchorCrewAgent: CrewAI agent registered with Anchor
- AnchorCrewMemory: Policy-checked shared memory with checkpoints

Usage:
    from anchor import Anchor
    from anchor.integrations.crewai import AnchorCrewAgent, AnchorCrewMemory
    from crewai import Crew

    anchor = Anchor(api_key="your-api-key")

    # Create agents with Anchor registration
    researcher = AnchorCrewAgent(
        anchor=anchor,
        role="Senior Researcher",
        goal="Find comprehensive information"
    )

    writer = AnchorCrewAgent(
        anchor=anchor,
        role="Content Writer",
        goal="Create engaging content"
    )

    # Create policy-checked shared memory
    memory = AnchorCrewMemory(anchor=anchor)

    # Configure policies
    anchor.config.update(memory.agent_id, {
        "policies": {"block_pii": True, "retention_days": 90}
    })

    crew = Crew(
        agents=[researcher.agent, writer.agent],
        tasks=[...],
        memory=memory
    )

    # Checkpoint & audit
    checkpoint_id = memory.create_checkpoint(label="before-experiment")
    events = anchor.audit.query(memory.agent_id, limit=50)
    verification = anchor.audit.verify(memory.agent_id)  # Verify chain integrity
"""

from typing import Any, List, Optional, Dict

# Check for CrewAI availability
try:
    from crewai import Agent

    CREWAI_AVAILABLE = True
except ImportError:
    CREWAI_AVAILABLE = False
    Agent = None


class AnchorCrewAgent:
    """
    CrewAI agent registered with Anchor.

    Description: Creates a CrewAI agent that is:
    - Registered in Anchor's agent registry
    - Has configurable policies (blocks PII, secrets, custom patterns)
    - Operations are audit-logged (hash-chained, tamper-evident)
    - Can query, verify, and export audit trails

    Usage:
        from anchor import Anchor
        from anchor.integrations.crewai import AnchorCrewAgent

        anchor = Anchor(api_key="your-api-key")

        researcher = AnchorCrewAgent(
            anchor=anchor,
            role="Senior Researcher",
            goal="Find comprehensive information",
            backstory="Expert researcher with 10 years experience"
        )

        # Configure policies for this agent
        anchor.config.update(researcher.agent_id, {
            "behavior": {"instructions": "Be thorough and accurate"},
            "policies": {"block_pii": True, "block_secrets": True}
        })

        # Access the CrewAI agent
        crew_agent = researcher.agent

        # Example: Query audit trail for this agent
        events = anchor.audit.query(researcher.agent_id, operations=["data.write"], limit=50)
        for event in events:
            print(f"{event.timestamp}: {event.operation} - {event.result}")

        # Example: Verify audit chain integrity
        verification = anchor.audit.verify(researcher.agent_id)
        print(f"Audit chain valid: {verification.valid}")
    """

    def __init__(
        self,
        anchor: Any,
        role: str,
        goal: str,
        backstory: Optional[str] = None,
        tools: Optional[List[Any]] = None,
        verbose: bool = False,
        **kwargs,
    ):
        """
        Initialize AnchorCrewAgent.

        Args:
            anchor: Anchor client instance
            role: Agent's role in the crew
            goal: Agent's goal
            backstory: Agent's backstory
            tools: List of tools for the agent
            verbose: Enable verbose output
            **kwargs: Additional CrewAI agent options
        """
        if not CREWAI_AVAILABLE:
            raise ImportError("CrewAI is required. Install with: pip install crewai")

        self.anchor = anchor
        self.role = role
        self.goal = goal

        # Register agent with Anchor
        self._anchor_agent = anchor.agents.create(
            name=f"crewai-{role.lower().replace(' ', '-')}",
            metadata={"framework": "crewai", "role": role, "goal": goal},
        )

        # Create CrewAI agent
        self._agent = Agent(
            role=role,
            goal=goal,
            backstory=backstory or f"An agent with the role of {role}",
            tools=tools or [],
            verbose=verbose,
            **kwargs,
        )

    @property
    def agent(self) -> Any:
        """Get the underlying CrewAI agent."""
        return self._agent

    @property
    def agent_id(self) -> str:
        """Get the Anchor agent ID."""
        return self._anchor_agent.id

    def store(
        self, key: str, value: str, metadata: Optional[Dict[str, Any]] = None
    ) -> bool:
        """
        Store data for this agent (policy-checked).

        Args:
            key: Storage key
            value: Value to store
            metadata: Optional metadata

        Returns:
            True if stored successfully
        """
        result = self.anchor.data.write(
            self._anchor_agent.id, key, value, metadata=metadata
        )
        return result.allowed

    def retrieve(self, key: str) -> Optional[str]:
        """
        Retrieve data for this agent.

        Args:
            key: Storage key

        Returns:
            Value or None if not found
        """
        return self.anchor.data.read(self._anchor_agent.id, key)

    def search(self, query: str, limit: int = 10) -> List[Dict[str, Any]]:
        """
        Search agent's data.

        Args:
            query: Search query
            limit: Maximum results

        Returns:
            List of matching entries
        """
        results = self.anchor.data.search(self._anchor_agent.id, query, limit=limit)
        return [
            {"key": r.key, "value": r.value, "similarity": r.similarity}
            for r in results
        ]


class AnchorCrewMemory:
    """
    Policy-checked shared memory for CrewAI with persistent storage and checkpoints.

    Description: Provides a memory backend for CrewAI that:
    - Enforces policies (PII blocking, etc.)
    - Logs all operations to audit trail
    - Supports checkpoints and rollback

    Usage:
        from anchor import Anchor
        from anchor.integrations.crewai import AnchorCrewMemory
        from crewai import Crew

        anchor = Anchor(api_key="your-api-key")
        memory = AnchorCrewMemory(anchor=anchor)
        anchor.config.update(memory.agent_id, {
            "policies": {"block_pii": True, "retention_days": 90}
        })
        crew = Crew(agents=[...], tasks=[...], memory=memory)
    """

    def __init__(self, anchor: Any, agent_id: Optional[str] = None):
        """
        Initialize AnchorCrewMemory.

        Args:
            anchor: Anchor client instance
            agent_id: Optional specific agent ID (creates one if not provided)
        """
        self.anchor = anchor

        # Create or use agent
        if agent_id:
            self._agent_id = agent_id
        else:
            agent = anchor.agents.create(
                "crewai-shared-memory",
                metadata={"framework": "crewai", "type": "shared-memory"},
            )
            self._agent_id = agent.id

    @property
    def agent_id(self) -> str:
        """Get the Anchor agent ID."""
        return self._agent_id

    def save(self, key: str, value: Any) -> bool:
        """
        Save a value to memory (policy-checked, persistent storage).

        Args:
            key: Memory key
            value: Value to store

        Returns:
            True if saved successfully (data persists across crew restarts)
        """
        # Convert value to string if needed
        str_value = value if isinstance(value, str) else str(value)

        result = self.anchor.data.write(
            self._agent_id, key, str_value, metadata={"source": "crewai"}
        )
        return result.allowed

    def search(self, query: str, limit: int = 10) -> List[Dict[str, Any]]:
        """
        Search memory.

        Args:
            query: Search query
            limit: Maximum results

        Returns:
            List of matching memories
        """
        results = self.anchor.data.search(self._agent_id, query, limit=limit)
        return [{"key": r.key, "value": r.value} for r in results]

    def get(self, key: str) -> Optional[Any]:
        """
        Get a value from memory.

        Args:
            key: Memory key

        Returns:
            Value or None if not found
        """
        return self.anchor.data.read(self._agent_id, key)

    def delete(self, key: str) -> bool:
        """
        Delete a value from memory.

        Args:
            key: Memory key

        Returns:
            True if deleted
        """
        return self.anchor.data.delete(self._agent_id, key)

    def clear(self) -> int:
        """
        Clear all memory.

        Returns:
            Number of keys deleted
        """
        return self.anchor.data.delete_prefix(self._agent_id, "")

    def create_checkpoint(self, label: Optional[str] = None) -> str:
        """
        Create a checkpoint of current memory state.

        Args:
            label: Optional label for the checkpoint

        Returns:
            Checkpoint ID

        Example:
            # Create checkpoint before risky operation
            checkpoint_id = memory.create_checkpoint(label="before-migration")
            
            # ... perform risky operations ...
            
            # If corruption detected, restore
            memory.restore_checkpoint(checkpoint_id)
        """
        checkpoint = self.anchor.checkpoints.create(
            self._agent_id, label=label or "crewai-checkpoint"
        )
        return checkpoint.id

    def restore_checkpoint(self, checkpoint_id: str) -> None:
        """
        Restore memory to a previous checkpoint.

        Args:
            checkpoint_id: ID of checkpoint to restore

        Example:
            # Restore to checkpoint after detecting corruption
            memory.restore_checkpoint(checkpoint_id)
            print("Memory restored to checkpoint")
        """
        self.anchor.checkpoints.restore(self._agent_id, checkpoint_id)



"""Base namespace class for Anchor SDK"""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .._http import HttpClient


class BaseNamespace:
    """Base class for all namespace classes"""

    def __init__(self, http: "HttpClient"):
        self._http = http



"""Tests for Anchor SDK with 5 namespaces (agents, config, data, checkpoints, audit)"""

import pytest
from unittest.mock import Mock, patch, MagicMock
from datetime import datetime

from anchor import Anchor, Agent, Config, ConfigVersion
from anchor import WriteResult, DataEntry, SearchResult
from anchor import Checkpoint, RestoreResult
from anchor import AuditEvent, Verification, ExportResult
from anchor.exceptions import NotFoundError, ValidationError


class TestAnchorClient:
    """Tests for the main Anchor client."""

    def test_initialization_with_api_key(self):
        """Test client initialization with API key."""
        anchor = Anchor(api_key="anc_test_key")
        assert anchor.client_config.api_key == "anc_test_key"
        assert anchor.client_config.base_url == "https://api.getanchor.dev"

    def test_initialization_requires_api_key(self):
        """Test that initialization fails without API key."""
        with pytest.raises(ValueError, match="api_key is required"):
            Anchor()

    def test_namespaces_available(self):
        """Test that all 5 namespaces are available."""
        anchor = Anchor(api_key="anc_test_key")
        assert hasattr(anchor, "agents")
        assert hasattr(anchor, "config")
        assert hasattr(anchor, "data")
        assert hasattr(anchor, "checkpoints")
        assert hasattr(anchor, "audit")


class TestAgentsNamespace:
    """Tests for agents namespace."""

    @pytest.fixture
    def anchor(self):
        return Anchor(api_key="anc_test_key")

    @pytest.fixture
    def mock_http(self, anchor):
        with patch.object(anchor._http, "post") as mock_post, patch.object(
            anchor._http, "get"
        ) as mock_get, patch.object(anchor._http, "patch") as mock_patch, patch.object(
            anchor._http, "delete"
        ) as mock_delete:
            yield {
                "post": mock_post,
                "get": mock_get,
                "patch": mock_patch,
                "delete": mock_delete,
            }

    def test_create_agent(self, anchor, mock_http):
        """Test creating an agent."""
        mock_http["post"].return_value = {
            "agent": {
                "id": "agent_123",
                "name": "test-agent",
                "status": "active",
                "metadata": {},
                "created_at": "2024-01-01T00:00:00Z",
            }
        }

        agent = anchor.agents.create("test-agent", {"env": "test"})

        assert isinstance(agent, Agent)
        assert agent.id == "agent_123"
        assert agent.name == "test-agent"
        assert agent.status == "active"

    def test_get_agent(self, anchor, mock_http):
        """Test getting an agent."""
        mock_http["get"].return_value = {
            "agent": {"id": "agent_123", "name": "test-agent", "status": "active"}
        }

        agent = anchor.agents.get("agent_123")

        assert agent.id == "agent_123"

    def test_list_agents(self, anchor, mock_http):
        """Test listing agents."""
        mock_http["get"].return_value = {
            "agents": [
                {"id": "agent_1", "name": "agent-1", "status": "active"},
                {"id": "agent_2", "name": "agent-2", "status": "suspended"},
            ]
        }

        agents = anchor.agents.list()

        assert len(agents) == 2
        assert agents[0].id == "agent_1"

    def test_suspend_agent(self, anchor, mock_http):
        """Test suspending an agent uses POST."""
        mock_http["post"].return_value = {
            "agent": {"id": "agent_123", "name": "test", "status": "suspended"}
        }

        agent = anchor.agents.suspend("agent_123")

        assert agent.status == "suspended"
        mock_http["post"].assert_called_once()

    def test_activate_agent(self, anchor, mock_http):
        """Test activating an agent uses POST."""
        mock_http["post"].return_value = {
            "agent": {"id": "agent_123", "name": "test", "status": "active"}
        }

        agent = anchor.agents.activate("agent_123")

        assert agent.status == "active"
        mock_http["post"].assert_called_once()


class TestConfigNamespace:
    """Tests for config namespace."""

    @pytest.fixture
    def anchor(self):
        return Anchor(api_key="anc_test_key")

    @pytest.fixture
    def mock_http(self, anchor):
        with patch.object(anchor._http, "get") as mock_get, patch.object(
            anchor._http, "put"
        ) as mock_put, patch.object(anchor._http, "post") as mock_post:
            yield {"get": mock_get, "put": mock_put, "post": mock_post}

    def test_get_config(self, anchor, mock_http):
        """Test getting config."""
        mock_http["get"].return_value = {
            "agent_id": "agent_123",
            "version": "v1",
            "config": {
                "instructions": "Be helpful",  # Freeform field
                "model": "gpt-4",
                "policies": {"block_pii": True},
            },
        }

        config = anchor.config.get("agent_123")

        assert isinstance(config, Config)
        assert config.version == "v1"
        assert config.config["instructions"] == "Be helpful"

    def test_update_config(self, anchor, mock_http):
        """Test updating config - schema-less with policies."""
        mock_http["put"].return_value = {
            "agent_id": "agent_123",
            "version": "v2",
            "config": {
                "role": "Support Agent",  # CrewAI-style
                "goal": "Help users",
                "policies": {"block_pii": True, "block_secrets": True}
            },
        }

        config = anchor.config.update(
            "agent_123", {
                "role": "Support Agent",
                "goal": "Help users",
                "policies": {"block_pii": True, "block_secrets": True}
            }
        )

        assert config.version == "v2"


class TestDataNamespace:
    """Tests for data namespace."""

    @pytest.fixture
    def anchor(self):
        return Anchor(api_key="anc_test_key")

    @pytest.fixture
    def mock_http(self, anchor):
        with patch.object(anchor._http, "post") as mock_post, patch.object(
            anchor._http, "get"
        ) as mock_get, patch.object(anchor._http, "delete") as mock_delete:
            yield {"post": mock_post, "get": mock_get, "delete": mock_delete}

    def test_write_data(self, anchor, mock_http):
        """Test writing data."""
        mock_http["post"].return_value = {
            "key": "user:123:pref",
            "allowed": True,
            "audit_id": "audit_abc",
        }

        result = anchor.data.write("agent_123", "user:123:pref", "dark_mode")

        assert isinstance(result, WriteResult)
        assert result.key == "user:123:pref"
        assert result.allowed is True

    def test_write_blocked_by_policy(self, anchor, mock_http):
        """Test write blocked by policy."""
        mock_http["post"].return_value = {
            "key": "user:123:ssn",
            "allowed": False,
            "blocked_by": "pii_filter",
            "reason": "SSN detected",
            "audit_id": "audit_def",
        }

        result = anchor.data.write("agent_123", "user:123:ssn", "123-45-6789")

        assert result.allowed is False
        assert result.blocked_by == "pii_filter"

    def test_read_data(self, anchor, mock_http):
        """Test reading data."""
        mock_http["get"].return_value = {"value": "dark_mode"}

        value = anchor.data.read("agent_123", "user:123:pref")

        assert value == "dark_mode"

    def test_read_not_found(self, anchor, mock_http):
        """Test reading non-existent data returns None."""
        mock_http["get"].side_effect = NotFoundError("Data", "key")

        value = anchor.data.read("agent_123", "nonexistent")

        assert value is None


class TestCheckpointsNamespace:
    """Tests for checkpoints namespace."""

    @pytest.fixture
    def anchor(self):
        return Anchor(api_key="anc_test_key")

    @pytest.fixture
    def mock_http(self, anchor):
        with patch.object(anchor._http, "post") as mock_post, patch.object(
            anchor._http, "get"
        ) as mock_get, patch.object(anchor._http, "delete") as mock_delete:
            yield {"post": mock_post, "get": mock_get, "delete": mock_delete}

    def test_create_checkpoint(self, anchor, mock_http):
        """Test creating a checkpoint."""
        mock_http["post"].return_value = {
            "id": "cp_123",
            "agent_id": "agent_123",
            "label": "v1.0",
            "config_version": "v5",
            "created_at": "2024-01-01T00:00:00Z",
        }

        checkpoint = anchor.checkpoints.create("agent_123", label="v1.0")

        assert isinstance(checkpoint, Checkpoint)
        assert checkpoint.id == "cp_123"
        assert checkpoint.label == "v1.0"

    def test_restore_checkpoint(self, anchor, mock_http):
        """Test restoring from a checkpoint."""
        mock_http["post"].return_value = {
            "restored_from": "cp_123",
            "config_restored": True,
            "config_version": "v5",
            "data_restored": True,
            "data_keys_restored": 50,
            "data_keys_removed": 10,
            "audit_id": "audit_xyz",
        }

        result = anchor.checkpoints.restore("agent_123", "cp_123")

        assert isinstance(result, RestoreResult)
        assert result.config_restored is True
        assert result.data_keys_restored == 50


class TestAuditNamespace:
    """Tests for audit namespace."""

    @pytest.fixture
    def anchor(self):
        return Anchor(api_key="anc_test_key")

    @pytest.fixture
    def mock_http(self, anchor):
        with patch.object(anchor._http, "get") as mock_get, patch.object(
            anchor._http, "post"
        ) as mock_post:
            yield {"get": mock_get, "post": mock_post}

    def test_query_audit_events(self, anchor, mock_http):
        """Test querying audit events."""
        mock_http["get"].return_value = {
            "events": [
                {
                    "id": "audit_1",
                    "agent_id": "agent_123",
                    "operation": "data.write",
                    "resource": "user:123:pref",
                    "result": "allowed",
                    "hash": "abc123",
                    "timestamp": "2024-01-01T00:00:00Z",
                }
            ]
        }

        events = anchor.audit.query("agent_123", operations=["data.write"])

        assert len(events) == 1
        assert isinstance(events[0], AuditEvent)
        assert events[0].operation == "data.write"

    def test_verify_chain(self, anchor, mock_http):
        """Test verifying audit chain."""
        mock_http["get"].return_value = {
            "valid": True,
            "events_checked": 100,
            "chain_start": "audit_001",
            "chain_end": "audit_100",
        }

        verification = anchor.audit.verify("agent_123")

        assert isinstance(verification, Verification)
        assert verification.valid is True
        assert verification.events_checked == 100

    def test_export_audit(self, anchor, mock_http):
        """Test exporting audit trail."""
        mock_http["post"].return_value = {
            "export_id": "exp_123",
            "format": "json",
            "download_url": "https://example.com/download",
            "event_count": 100,
        }

        result = anchor.audit.export("agent_123", format="json")

        assert isinstance(result, ExportResult)
        assert result.format == "json"
        assert result.event_count == 100



"""Checkpoints namespace for snapshot and rollback operations."""

from typing import Optional, Dict, Any, List
from dataclasses import dataclass, field
from datetime import datetime

from .base import BaseNamespace


@dataclass
class Checkpoint:
    """Checkpoint snapshot."""

    id: str
    agent_id: str
    label: Optional[str] = None
    description: Optional[str] = None
    config_version: str = ""
    data_snapshot: Dict[str, Any] = field(default_factory=dict)
    created_at: Optional[datetime] = None
    audit_id: Optional[str] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Checkpoint":
        return cls(
            id=data.get("id", ""),
            agent_id=data.get("agent_id", ""),
            label=data.get("label"),
            description=data.get("description"),
            config_version=data.get("config_version", ""),
            data_snapshot=data.get("data_snapshot", {}),
            created_at=(
                datetime.fromisoformat(data["created_at"].replace("Z", "+00:00"))
                if data.get("created_at")
                else None
            ),
            audit_id=data.get("audit_id"),
        )


@dataclass
class RestoreResult:
    """Result of a checkpoint restore operation."""

    restored_from: str
    config_restored: bool
    config_version: Optional[str] = None
    data_restored: bool = False
    data_keys_restored: int = 0
    data_keys_removed: int = 0
    audit_id: str = ""
    restored_at: Optional[datetime] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "RestoreResult":
        return cls(
            restored_from=data.get("restored_from", ""),
            config_restored=data.get("config_restored", False),
            config_version=data.get("config_version"),
            data_restored=data.get("data_restored", False),
            data_keys_restored=data.get("data_keys_restored", 0),
            data_keys_removed=data.get("data_keys_removed", 0),
            audit_id=data.get("audit_id", ""),
            restored_at=(
                datetime.fromisoformat(data["restored_at"].replace("Z", "+00:00"))
                if data.get("restored_at")
                else None
            ),
        )


class CheckpointsNamespace(BaseNamespace):
    """Checkpoint management for rollback."""

    def create(
        self,
        agent_id: str,
        label: Optional[str] = None,
        description: Optional[str] = None,
    ) -> Checkpoint:
        """
        Create a checkpoint of current state.

        Args:
            agent_id: Agent ID
            label: Optional human-readable label
            description: Optional description

        Returns:
            Checkpoint object
        """
        response = self._http.post(
            f"/agents/{agent_id}/checkpoints",
            data={
                "label": label,
                "description": description,
            },
        )
        return Checkpoint.from_dict(response)

    def list(self, agent_id: str, limit: int = 20) -> List[Checkpoint]:
        """
        List checkpoints.

        Args:
            agent_id: Agent ID
            limit: Max checkpoints to return

        Returns:
            List of Checkpoint objects
        """
        response = self._http.get(
            f"/agents/{agent_id}/checkpoints", params={"limit": limit}
        )
        return [Checkpoint.from_dict(c) for c in response.get("checkpoints", [])]

    def get(self, agent_id: str, checkpoint_id: str) -> Checkpoint:
        """
        Get a specific checkpoint.

        Args:
            agent_id: Agent ID
            checkpoint_id: Checkpoint ID

        Returns:
            Checkpoint object
        """
        response = self._http.get(f"/agents/{agent_id}/checkpoints/{checkpoint_id}")
        return Checkpoint.from_dict(response)

    def restore(
        self,
        agent_id: str,
        checkpoint_id: str,
        restore_config: bool = True,
        restore_data: bool = True,
    ) -> RestoreResult:
        """
        Restore agent to a checkpoint.

        Args:
            agent_id: Agent ID
            checkpoint_id: Checkpoint to restore
            restore_config: Whether to restore config
            restore_data: Whether to restore data

        Returns:
            RestoreResult with details
        """
        response = self._http.post(
            f"/agents/{agent_id}/checkpoints/{checkpoint_id}/restore",
            data={
                "restore_config": restore_config,
                "restore_data": restore_data,
            },
        )
        return RestoreResult.from_dict(response)

    def delete(self, agent_id: str, checkpoint_id: str) -> bool:
        """
        Delete a checkpoint.

        Args:
            agent_id: Agent ID
            checkpoint_id: Checkpoint ID

        Returns:
            True if deleted
        """
        self._http.delete(f"/agents/{agent_id}/checkpoints/{checkpoint_id}")
        return True



"""HTTP client wrapper for Anchor SDK."""

import os
from typing import Any, Dict, Optional
import httpx

from anchor.exceptions import (
    AnchorError,
    AuthenticationError,
    NotFoundError,
    RateLimitError,
    ValidationError,
)


class HTTPClient:
    """HTTP client for Anchor API."""

    def __init__(
        self,
        api_key: Optional[str] = None,
        base_url: str = "https://api.getanchor.dev",
        timeout: float = 30.0,
    ):
        self.api_key = api_key or os.environ.get("ANCHOR_API_KEY")
        if not self.api_key:
            raise AuthenticationError(
                "API key required. Pass api_key or set ANCHOR_API_KEY."
            )

        self.base_url = base_url.rstrip("/")
        self.timeout = timeout
        self._client = httpx.Client(timeout=timeout)

    def _headers(self) -> Dict[str, str]:
        return {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }

    def _handle_response(self, response: httpx.Response) -> Dict[str, Any]:
        """Handle API response and raise appropriate errors."""
        if response.status_code == 401:
            raise AuthenticationError("Invalid API key")

        if response.status_code == 404:
            data = response.json() if response.content else {}
            raise NotFoundError(
                data.get("resource_type", "Resource"),
                data.get("resource_id", "unknown"),
            )

        if response.status_code == 429:
            retry_after = int(response.headers.get("Retry-After", 60))
            raise RateLimitError(retry_after)

        if response.status_code == 422:
            data = response.json() if response.content else {}
            raise ValidationError(
                data.get("message", "Validation error"),
                data.get("field"),
            )

        if response.status_code >= 400:
            data = response.json() if response.content else {}
            raise AnchorError(
                data.get("message", f"API error: {response.status_code}"),
                data.get("code"),
            )

        if response.status_code == 204:
            return {}

        return response.json()

    def get(
        self,
        path: str,
        params: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Make GET request."""
        url = f"{self.base_url}{path}"
        response = self._client.get(url, headers=self._headers(), params=params)
        return self._handle_response(response)

    def post(
        self,
        path: str,
        json: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Make POST request."""
        url = f"{self.base_url}{path}"
        response = self._client.post(url, headers=self._headers(), json=json)
        return self._handle_response(response)

    def put(
        self,
        path: str,
        json: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Make PUT request."""
        url = f"{self.base_url}{path}"
        response = self._client.put(url, headers=self._headers(), json=json)
        return self._handle_response(response)

    def patch(
        self,
        path: str,
        json: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Make PATCH request."""
        url = f"{self.base_url}{path}"
        response = self._client.patch(url, headers=self._headers(), json=json)
        return self._handle_response(response)

    def delete(
        self,
        path: str,
        params: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Make DELETE request."""
        url = f"{self.base_url}{path}"
        response = self._client.delete(url, headers=self._headers(), params=params)
        return self._handle_response(response)

    def close(self):
        """Close the HTTP client."""
        self._client.close()



"""
Anchor: Control what your AI agents store. Audit everything.

Anchor lets you persist agent state, block bad data, prove what happened.

1. Quick Start:
    from anchor import Anchor

    anchor = Anchor(api_key="your-api-key")
    agent = anchor.agents.create("my-agent")

2. Full Example:
    from anchor import Anchor

    anchor = Anchor(api_key="your-api-key")

    # Create agent
    agent = anchor.agents.create(
        name="support-bot",
        metadata={"environment": "production"}
    )

    # Store data (automatically audit-logged)
    anchor.data.write(agent.id, "user:123:pref", "concise answers")
    print(result.allowed)  # True

    # PII gets blocked by policy
    result = anchor.data.write(agent.id, "user:123:ssn", "123-45-6789")
    print(result.allowed)  # False - blocked by policy

    # Checkpoint before risky operation
    checkpoint = anchor.checkpoints.create(agent.id, label="pre-update")

    # Something goes wrong? Rollback
    anchor.checkpoints.restore(agent.id, checkpoint.id)

    # Prove what happened
    events = anchor.audit.query(agent.id, limit=5)
        from anchor import Anchor
"""

from typing import Optional

from .config import Config as ClientConfig
from ._http import HttpClient
from .namespaces import (
    AgentsNamespace,
    ConfigNamespace,
    DataNamespace,
    CheckpointsNamespace,
    AuditNamespace,
)


class Anchor:
    """
    Anchor client for AI agent state management.

    Usage:
        # Simple initialization
        anchor = Anchor(api_key="your-api-key")

        # With configuration
        from anchorai import Anchor, ClientConfig

        config = ClientConfig(
            api_key="your-api-key",
            base_url="https://api.getanchor.dev",
            timeout=30.0,
            retry_attempts=3
        )
        anchor = Anchor(config=config)

    Namespaces:
        anchor.agents       - Agent registry and lifecycle
        anchor.config       - Agent configuration with versioning
        anchor.data         - Governed key-value data storage
        anchor.checkpoints  - State snapshots and rollback
        anchor.audit        - Hash-chained audit trail
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        config: Optional[ClientConfig] = None,
        **kwargs,
    ):
        """
        Initialize Anchor client.

        Args:
            api_key: API key
            base_url: API base URL (default: https://api.getanchor.dev)
            config: Full configuration object
            **kwargs: Additional configuration options passed to Config
        """
        if config is not None:
            self._config = config
        else:
            if api_key is None:
                raise ValueError("api_key is required")

            self._config = ClientConfig(
                api_key=api_key,
                base_url=base_url or "https://api.getanchor.dev",
                **kwargs,
            )

        # Initialize HTTP client
        self._http = HttpClient(self._config)

        # Initialize namespaces (5 per spec)
        self._agents = AgentsNamespace(self._http)
        self._config_ns = ConfigNamespace(self._http)
        self._data = DataNamespace(self._http)
        self._checkpoints = CheckpointsNamespace(self._http)
        self._audit = AuditNamespace(self._http)

    @property
    def agents(self) -> AgentsNamespace:
        """Agent registry and lifecycle management."""
        return self._agents

    @property
    def config(self) -> ConfigNamespace:
        """Agent configuration with versioning."""
        return self._config_ns

    @property
    def data(self) -> DataNamespace:
        """Governed key-value data storage."""
        return self._data

    @property
    def checkpoints(self) -> CheckpointsNamespace:
        """State snapshots and rollback."""
        return self._checkpoints

    @property
    def audit(self) -> AuditNamespace:
        """Hash-chained audit trail."""
        return self._audit

    @property
    def client_config(self) -> ClientConfig:
        """Current client configuration."""
        return self._config

    def __repr__(self) -> str:
        return f"Anchor(base_url='{self._config.base_url}')"



"""Configuration for Anchor SDK."""

from dataclasses import dataclass
from typing import Optional


@dataclass
class Config:
    """
    Configuration for the Anchor client.

    Usage:
        from anchorai import Anchor, Config

        config = Config(
            api_key="your-api-key",
            base_url="https://api.getanchor.dev",
            timeout=30.0,
            retry_attempts=3
        )
        anchor = Anchor(config=config)
    """

    api_key: str
    base_url: str = "https://api.getanchor.dev"
    region: Optional[str] = None
    timeout: float = 30.0
    retry_attempts: int = 3
    retry_delay: float = 1.0
    batch_size: int = 100
    cache_policies: bool = True
    verify_ssl: bool = True
    log_level: str = "INFO"

    def __post_init__(self):
        """Validate and normalize configuration."""
        self.base_url = self.base_url.rstrip("/")

        # Handle regional endpoints
        if self.region and "api.getanchor.dev" in self.base_url:
            self.base_url = f"https://{self.region}.api.getanchor.dev"



"""Tests for framework integrations."""

import pytest
from unittest.mock import Mock, patch, MagicMock
from datetime import datetime, timezone

from anchor import Anchor
from anchor.models import WriteResult, DataEntry


class TestIntegrationsInit:
    """Tests for integrations __init__.py lazy loading."""

    def test_lazy_import_langchain(self):
        """Test that langchain integration can be imported."""
        # Direct import should work
        from anchor.integrations.langchain import AnchorMemory, AnchorChatHistory

        assert AnchorMemory is not None
        assert AnchorChatHistory is not None

    def test_lazy_import_crewai(self):
        """Test that crewai integration can be imported."""
        # Direct import should work
        from anchor.integrations.crewai import AnchorCrewAgent, AnchorCrewMemory

        assert AnchorCrewAgent is not None
        assert AnchorCrewMemory is not None

    def test_lazy_import_mem0(self):
        """Test that mem0 integration can be imported."""
        # Direct import should work
        from anchor.integrations.mem0 import AnchorMem0

        assert AnchorMem0 is not None

    def test_integrations_module_has_getattr(self):
        """Test that integrations module has __getattr__ for lazy loading."""
        import anchor.integrations as integrations

        # Test that __getattr__ exists (implementation detail)
        assert hasattr(integrations, "__getattr__")


class TestLangChainIntegration:
    """Tests for LangChain integration."""

    @pytest.fixture
    def anchor(self):
        """Create a mock Anchor client."""
        anchor = Mock(spec=Anchor)
        anchor.agents = Mock()
        anchor.data = Mock()
        anchor.checkpoints = Mock()
        anchor.config = Mock()
        return anchor

    @pytest.fixture
    def agent_id(self):
        """Return a test agent ID."""
        return "agent_123"

    @patch("anchor.integrations.langchain.LANGCHAIN_AVAILABLE", True)
    @patch("anchor.integrations.langchain.HumanMessage")
    @patch("anchor.integrations.langchain.AIMessage")
    @patch("anchor.integrations.langchain.SystemMessage")
    def test_anchor_memory_initialization(
        self, mock_system, mock_ai, mock_human, anchor, agent_id
    ):
        """Test AnchorMemory initialization."""
        from anchor.integrations.langchain import AnchorMemory

        memory = AnchorMemory(anchor=anchor, agent_id=agent_id)
        assert memory.anchor == anchor
        assert memory.agent_id == agent_id
        assert memory.memory_key == "chat_history"
        assert memory.return_messages is True
        assert memory.session_id is not None

    @patch("anchor.integrations.langchain.LANGCHAIN_AVAILABLE", True)
    def test_anchor_memory_custom_params(self, anchor, agent_id):
        """Test AnchorMemory with custom parameters."""
        from anchor.integrations.langchain import AnchorMemory

        memory = AnchorMemory(
            anchor=anchor,
            agent_id=agent_id,
            memory_key="history",
            return_messages=False,
            session_id="custom_session",
        )
        assert memory.memory_key == "history"
        assert memory.return_messages is False
        assert memory.session_id == "custom_session"

    @patch("anchor.integrations.langchain.LANGCHAIN_AVAILABLE", True)
    def test_anchor_memory_memory_variables(self, anchor, agent_id):
        """Test memory_variables property."""
        from anchor.integrations.langchain import AnchorMemory

        memory = AnchorMemory(anchor=anchor, agent_id=agent_id, memory_key="history")
        assert memory.memory_variables == ["history"]

    @patch("anchor.integrations.langchain.LANGCHAIN_AVAILABLE", True)
    @patch("anchor.integrations.langchain.HumanMessage")
    @patch("anchor.integrations.langchain.AIMessage")
    def test_anchor_memory_save_context(
        self, mock_ai, mock_human, anchor, agent_id
    ):
        """Test saving context to Anchor."""
        from anchor.integrations.langchain import AnchorMemory

        # Mock write results
        anchor.data.write.return_value = WriteResult(
            key="test", allowed=True, audit_id="audit_123"
        )

        memory = AnchorMemory(anchor=anchor, agent_id=agent_id, session_id="test_session")
        memory.save_context(
            inputs={"input": "Hello"}, outputs={"output": "Hi there"}
        )

        # Should have called write twice (for input and output)
        assert anchor.data.write.call_count == 2

    @patch("anchor.integrations.langchain.LANGCHAIN_AVAILABLE", True)
    @patch("anchor.integrations.langchain.HumanMessage")
    @patch("anchor.integrations.langchain.AIMessage")
    def test_anchor_memory_load_memory_variables(
        self, mock_ai, mock_human, anchor, agent_id
    ):
        """Test loading memory variables."""
        from anchor.integrations.langchain import AnchorMemory
        from anchor.models import DataEntry

        # Mock list and read_full
        anchor.data.list.return_value = ["chat:test:human:1", "chat:test:ai:2"]

        # Create mock entries
        human_entry = DataEntry(
            key="chat:test:human:1",
            value="Hello",
            metadata={"role": "human"},
            created_at=datetime.now(timezone.utc),
            updated_at=datetime.now(timezone.utc),
            audit_id="audit_123",
        )
        ai_entry = DataEntry(
            key="chat:test:ai:2",
            value="Hi there",
            metadata={"role": "ai"},
            created_at=datetime.now(timezone.utc),
            updated_at=datetime.now(timezone.utc),
            audit_id="audit_123",
        )

        def read_full_side_effect(agent_id, key):
            if "human" in key:
                return human_entry
            return ai_entry

        anchor.data.read_full.side_effect = read_full_side_effect

        memory = AnchorMemory(anchor=anchor, agent_id=agent_id, session_id="test")
        result = memory.load_memory_variables({})

        assert memory.memory_key in result
        anchor.data.list.assert_called_once()

    @patch("anchor.integrations.langchain.LANGCHAIN_AVAILABLE", True)
    def test_anchor_memory_clear(self, anchor, agent_id):
        """Test clearing memory."""
        from anchor.integrations.langchain import AnchorMemory

        memory = AnchorMemory(anchor=anchor, agent_id=agent_id, session_id="test")
        memory.clear()

        assert len(memory._messages) == 0
        anchor.data.delete_prefix.assert_called_once_with(
            agent_id, "chat:test:"
        )

    @patch("anchor.integrations.langchain.LANGCHAIN_AVAILABLE", True)
    def test_anchor_memory_create_checkpoint(self, anchor, agent_id):
        """Test creating checkpoint."""
        from anchor.integrations.langchain import AnchorMemory
        from anchor.models import Checkpoint, DataSnapshot

        mock_checkpoint = Checkpoint(
            id="checkpoint_123",
            agent_id=agent_id,
            config_version="v1",
            data_snapshot=DataSnapshot(key_count=10, total_size_bytes=1024),
            created_at=datetime.now(timezone.utc),
        )
        anchor.checkpoints.create.return_value = mock_checkpoint

        memory = AnchorMemory(anchor=anchor, agent_id=agent_id, session_id="test")
        checkpoint_id = memory.create_checkpoint(label="test-checkpoint")

        assert checkpoint_id == "checkpoint_123"
        anchor.checkpoints.create.assert_called_once()

    @patch("anchor.integrations.langchain.LANGCHAIN_AVAILABLE", True)
    @patch("anchor.integrations.langchain.AnchorMemory._load_from_anchor")
    def test_anchor_memory_restore_checkpoint(self, mock_load, anchor, agent_id):
        """Test restoring checkpoint."""
        from anchor.integrations.langchain import AnchorMemory

        memory = AnchorMemory(anchor=anchor, agent_id=agent_id, session_id="test")
        memory.restore_checkpoint("checkpoint_123")

        anchor.checkpoints.restore.assert_called_once_with(agent_id, "checkpoint_123")
        mock_load.assert_called_once()

    @patch("anchor.integrations.langchain.LANGCHAIN_AVAILABLE", False)
    def test_anchor_memory_without_langchain(self, anchor, agent_id):
        """Test AnchorMemory when LangChain is not available."""
        from anchor.integrations.langchain import AnchorMemory

        # Should still work, just won't use LangChain message types
        memory = AnchorMemory(anchor=anchor, agent_id=agent_id)
        assert memory.anchor == anchor

    @patch("anchor.integrations.langchain.LANGCHAIN_AVAILABLE", True)
    @patch("anchor.integrations.langchain.BaseChatMessageHistory")
    def test_anchor_chat_history_initialization(
        self, mock_base, anchor, agent_id
    ):
        """Test AnchorChatHistory initialization."""
        from anchor.integrations.langchain import AnchorChatHistory

        history = AnchorChatHistory(anchor=anchor, agent_id=agent_id)
        assert history.anchor == anchor
        assert history.agent_id == agent_id

    @patch("anchor.integrations.langchain.LANGCHAIN_AVAILABLE", False)
    def test_anchor_chat_history_without_langchain(self, anchor, agent_id):
        """Test AnchorChatHistory when LangChain is not available."""
        from anchor.integrations.langchain import AnchorChatHistory

        with pytest.raises(ImportError):
            AnchorChatHistory(anchor=anchor, agent_id=agent_id)


class TestCrewAIIntegration:
    """Tests for CrewAI integration."""

    @pytest.fixture
    def anchor(self):
        """Create a mock Anchor client."""
        anchor = Mock(spec=Anchor)
        anchor.agents = Mock()
        anchor.data = Mock()
        anchor.config = Mock()
        return anchor

    @patch("anchor.integrations.crewai.CREWAI_AVAILABLE", True)
    @patch("anchor.integrations.crewai.Agent")
    def test_anchor_crew_agent_initialization(self, mock_agent_class, anchor):
        """Test AnchorCrewAgent initialization."""
        from anchor.integrations.crewai import AnchorCrewAgent

        # Mock agent creation
        mock_anchor_agent = Mock()
        mock_anchor_agent.id = "agent_123"
        anchor.agents.create.return_value = mock_anchor_agent

        # Mock CrewAI agent
        mock_crew_agent = Mock()
        mock_agent_class.return_value = mock_crew_agent

        crew_agent = AnchorCrewAgent(
            anchor=anchor, role="Researcher", goal="Research topics"
        )

        assert crew_agent.anchor == anchor
        assert crew_agent.role == "Researcher"
        assert crew_agent.goal == "Research topics"
        assert crew_agent.agent_id == "agent_123"
        assert crew_agent.agent == mock_crew_agent

    @patch("anchor.integrations.crewai.CREWAI_AVAILABLE", True)
    @patch("anchor.integrations.crewai.Agent")
    def test_anchor_crew_agent_store(self, mock_agent_class, anchor):
        """Test storing data."""
        from anchor.integrations.crewai import AnchorCrewAgent

        mock_anchor_agent = Mock()
        mock_anchor_agent.id = "agent_123"
        anchor.agents.create.return_value = mock_anchor_agent
        mock_agent_class.return_value = Mock()

        anchor.data.write.return_value = WriteResult(
            key="test", allowed=True, audit_id="audit_123"
        )

        crew_agent = AnchorCrewAgent(
            anchor=anchor, role="Researcher", goal="Research"
        )
        result = crew_agent.store("key1", "value1", metadata={"source": "test"})

        assert result is True
        anchor.data.write.assert_called_once()

    @patch("anchor.integrations.crewai.CREWAI_AVAILABLE", True)
    @patch("anchor.integrations.crewai.Agent")
    def test_anchor_crew_agent_retrieve(self, mock_agent_class, anchor):
        """Test retrieving data."""
        from anchor.integrations.crewai import AnchorCrewAgent

        mock_anchor_agent = Mock()
        mock_anchor_agent.id = "agent_123"
        anchor.agents.create.return_value = mock_anchor_agent
        mock_agent_class.return_value = Mock()

        anchor.data.read.return_value = "stored_value"

        crew_agent = AnchorCrewAgent(anchor=anchor, role="Researcher", goal="Research")
        result = crew_agent.retrieve("key1")

        assert result == "stored_value"
        anchor.data.read.assert_called_once_with("agent_123", "key1")

    @patch("anchor.integrations.crewai.CREWAI_AVAILABLE", True)
    @patch("anchor.integrations.crewai.Agent")
    def test_anchor_crew_agent_search(self, mock_agent_class, anchor):
        """Test searching data."""
        from anchor.integrations.crewai import AnchorCrewAgent
        from anchor.models import SearchResult

        mock_anchor_agent = Mock()
        mock_anchor_agent.id = "agent_123"
        anchor.agents.create.return_value = mock_anchor_agent
        mock_agent_class.return_value = Mock()

        mock_results = [
            SearchResult(key="k1", value="v1", similarity=0.9),
            SearchResult(key="k2", value="v2", similarity=0.8),
        ]
        anchor.data.search.return_value = mock_results

        crew_agent = AnchorCrewAgent(anchor=anchor, role="Researcher", goal="Research")
        results = crew_agent.search("query", limit=5)

        assert len(results) == 2
        assert results[0]["key"] == "k1"
        assert results[0]["similarity"] == 0.9

    @patch("anchor.integrations.crewai.CREWAI_AVAILABLE", True)
    def test_anchor_crew_memory_initialization(self, anchor):
        """Test AnchorCrewMemory initialization."""
        from anchor.integrations.crewai import AnchorCrewMemory

        mock_agent = Mock()
        mock_agent.id = "agent_123"
        anchor.agents.create.return_value = mock_agent

        memory = AnchorCrewMemory(anchor=anchor)
        assert memory.anchor == anchor
        assert memory.agent_id == "agent_123"

    @patch("anchor.integrations.crewai.CREWAI_AVAILABLE", True)
    def test_anchor_crew_memory_with_agent_id(self, anchor):
        """Test AnchorCrewMemory with existing agent ID."""
        from anchor.integrations.crewai import AnchorCrewMemory

        memory = AnchorCrewMemory(anchor=anchor, agent_id="existing_agent")
        assert memory.agent_id == "existing_agent"
        anchor.agents.create.assert_not_called()

    @patch("anchor.integrations.crewai.CREWAI_AVAILABLE", True)
    def test_anchor_crew_memory_save(self, anchor):
        """Test saving to memory."""
        from anchor.integrations.crewai import AnchorCrewMemory

        mock_agent = Mock()
        mock_agent.id = "agent_123"
        anchor.agents.create.return_value = mock_agent

        anchor.data.write.return_value = WriteResult(
            key="test", allowed=True, audit_id="audit_123"
        )

        memory = AnchorCrewMemory(anchor=anchor)
        result = memory.save("key1", "value1")

        assert result is True
        anchor.data.write.assert_called_once()

    @patch("anchor.integrations.crewai.CREWAI_AVAILABLE", False)
    def test_anchor_crew_agent_without_crewai(self, anchor):
        """Test AnchorCrewAgent when CrewAI is not available."""
        from anchor.integrations.crewai import AnchorCrewAgent

        with pytest.raises(ImportError, match="CrewAI is required"):
            AnchorCrewAgent(anchor=anchor, role="Researcher", goal="Research")


class TestMem0Integration:
    """Tests for Mem0 integration."""

    @pytest.fixture
    def anchor(self):
        """Create a mock Anchor client."""
        anchor = Mock(spec=Anchor)
        anchor.data = Mock()
        return anchor

    @pytest.fixture
    def agent_id(self):
        """Return a test agent ID."""
        return "agent_123"

    @pytest.fixture
    def mock_mem0(self):
        """Create a mock Mem0 client."""
        mem0 = Mock()
        mem0.add = Mock(return_value={"id": "mem_123"})
        mem0.search = Mock(return_value=[{"id": "mem_123", "content": "test"}])
        mem0.get = Mock(return_value={"id": "mem_123", "content": "test"})
        mem0.get_all = Mock(return_value=[{"id": "mem_123"}])
        mem0.update = Mock(return_value={"id": "mem_123"})
        mem0.delete = Mock(return_value=True)
        return mem0

    @patch("anchor.integrations.mem0.MEM0_AVAILABLE", True)
    def test_anchor_mem0_initialization(self, anchor, agent_id, mock_mem0):
        """Test AnchorMem0 initialization."""
        from anchor.integrations.mem0 import AnchorMem0

        wrapped = AnchorMem0(anchor=anchor, agent_id=agent_id, mem0_client=mock_mem0)
        assert wrapped.anchor == anchor
        assert wrapped.agent_id == agent_id
        assert wrapped.mem0 == mock_mem0

    @patch("anchor.integrations.mem0.MEM0_AVAILABLE", True)
    def test_anchor_mem0_add_allowed(self, anchor, agent_id, mock_mem0):
        """Test adding memory when allowed."""
        from anchor.integrations.mem0 import AnchorMem0

        anchor.data.write.return_value = WriteResult(
            key="test", allowed=True, audit_id="audit_123"
        )

        wrapped = AnchorMem0(anchor=anchor, agent_id=agent_id, mem0_client=mock_mem0)
        result = wrapped.add("User prefers dark mode", user_id="user_123")

        assert result.allowed is True
        assert result.blocked_by is None
        assert result.mem0_result == {"id": "mem_123"}
        mock_mem0.add.assert_called_once()

    @patch("anchor.integrations.mem0.MEM0_AVAILABLE", True)
    def test_anchor_mem0_add_blocked(self, anchor, agent_id, mock_mem0):
        """Test adding memory when blocked by policy."""
        from anchor.integrations.mem0 import AnchorMem0

        anchor.data.write.return_value = WriteResult(
            key="test",
            allowed=False,
            audit_id="audit_123",
            blocked_by="pii_filter",
            reason="Email detected",
        )

        wrapped = AnchorMem0(anchor=anchor, agent_id=agent_id, mem0_client=mock_mem0)
        result = wrapped.add("user@example.com", user_id="user_123")

        assert result.allowed is False
        assert result.blocked_by == "pii_filter"
        assert result.reason == "Email detected"
        assert result.mem0_result is None
        mock_mem0.add.assert_not_called()

    @patch("anchor.integrations.mem0.MEM0_AVAILABLE", True)
    def test_anchor_mem0_search(self, anchor, agent_id, mock_mem0):
        """Test searching memories."""
        from anchor.integrations.mem0 import AnchorMem0

        anchor.data.write.return_value = WriteResult(
            key="test", allowed=True, audit_id="audit_123"
        )

        wrapped = AnchorMem0(anchor=anchor, agent_id=agent_id, mem0_client=mock_mem0)
        results = wrapped.search("dark mode", user_id="user_123", limit=5)

        assert len(results) == 1
        assert results[0]["id"] == "mem_123"
        mock_mem0.search.assert_called_once()
        # Should have logged the search
        anchor.data.write.assert_called_once()

    @patch("anchor.integrations.mem0.MEM0_AVAILABLE", True)
    def test_anchor_mem0_get(self, anchor, agent_id, mock_mem0):
        """Test getting a memory by ID."""
        from anchor.integrations.mem0 import AnchorMem0

        wrapped = AnchorMem0(anchor=anchor, agent_id=agent_id, mem0_client=mock_mem0)
        result = wrapped.get("mem_123")

        assert result == {"id": "mem_123", "content": "test"}
        mock_mem0.get.assert_called_once_with("mem_123")

    @patch("anchor.integrations.mem0.MEM0_AVAILABLE", True)
    def test_anchor_mem0_get_all(self, anchor, agent_id, mock_mem0):
        """Test getting all memories."""
        from anchor.integrations.mem0 import AnchorMem0

        wrapped = AnchorMem0(anchor=anchor, agent_id=agent_id, mem0_client=mock_mem0)
        results = wrapped.get_all(user_id="user_123")

        assert len(results) == 1
        mock_mem0.get_all.assert_called_once()

    @patch("anchor.integrations.mem0.MEM0_AVAILABLE", True)
    def test_anchor_mem0_update_allowed(self, anchor, agent_id, mock_mem0):
        """Test updating memory when allowed."""
        from anchor.integrations.mem0 import AnchorMem0

        anchor.data.write.return_value = WriteResult(
            key="test", allowed=True, audit_id="audit_123"
        )

        wrapped = AnchorMem0(anchor=anchor, agent_id=agent_id, mem0_client=mock_mem0)
        result = wrapped.update("mem_123", "Updated content")

        assert result.allowed is True
        mock_mem0.update.assert_called_once()

    @patch("anchor.integrations.mem0.MEM0_AVAILABLE", True)
    def test_anchor_mem0_update_blocked(self, anchor, agent_id, mock_mem0):
        """Test updating memory when blocked."""
        from anchor.integrations.mem0 import AnchorMem0

        anchor.data.write.return_value = WriteResult(
            key="test", allowed=False, audit_id="audit_123", blocked_by="pii_filter"
        )

        wrapped = AnchorMem0(anchor=anchor, agent_id=agent_id, mem0_client=mock_mem0)
        result = wrapped.update("mem_123", "user@example.com")

        assert result.allowed is False
        mock_mem0.update.assert_not_called()

    @patch("anchor.integrations.mem0.MEM0_AVAILABLE", True)
    def test_anchor_mem0_delete(self, anchor, agent_id, mock_mem0):
        """Test deleting a memory."""
        from anchor.integrations.mem0 import AnchorMem0

        wrapped = AnchorMem0(anchor=anchor, agent_id=agent_id, mem0_client=mock_mem0)
        result = wrapped.delete("mem_123")

        assert result is True
        mock_mem0.delete.assert_called_once_with("mem_123")

    @patch("anchor.integrations.mem0.MEM0_AVAILABLE", False)
    def test_anchor_mem0_without_mem0(self, anchor, agent_id):
        """Test AnchorMem0 when Mem0 is not available."""
        from anchor.integrations.mem0 import AnchorMem0

        with pytest.raises(ImportError, match="Mem0 is required"):
            AnchorMem0(anchor=anchor, agent_id=agent_id, mem0_client=Mock())



"""Namespace classes for Anchor SDK."""

from .agents import AgentsNamespace, Agent
from .config import ConfigNamespace, Config, ConfigVersion
from .data import DataNamespace, WriteResult, DataEntry, SearchResult
from .checkpoints import CheckpointsNamespace, Checkpoint, RestoreResult
from .audit import AuditNamespace, AuditEvent, Verification, ExportResult

__all__ = [
    # Namespaces
    "AgentsNamespace",
    "ConfigNamespace",
    "DataNamespace",
    "CheckpointsNamespace",
    "AuditNamespace",
    # Agent types
    "Agent",
    # Config types
    "Config",
    "ConfigVersion",
    # Data types
    "WriteResult",
    "DataEntry",
    "SearchResult",
    # Checkpoint types
    "Checkpoint",
    "RestoreResult",
    # Audit types
    "AuditEvent",
    "Verification",
    "ExportResult",
]



"""Tests for HTTP client utilities."""

import pytest
import time
from unittest.mock import Mock, patch, MagicMock
import requests
from requests.exceptions import Timeout, ConnectionError, RequestException

from anchor.config import Config
from anchor._http import HttpClient
from anchor.exceptions import (
    AnchorAPIError,
    AuthenticationError,
    AuthorizationError,
    NotFoundError,
    ValidationError,
    RateLimitError,
    ServerError,
    NetworkError,
    PolicyViolationError,
)


class TestHttpClient:
    """Tests for HttpClient class."""

    @pytest.fixture
    def config(self):
        """Create a test config."""
        return Config(
            api_key="anc_test_key",
            base_url="https://api.getanchor.dev",
            timeout=30.0,
            retry_attempts=3,
            retry_delay=0.1,
        )

    @pytest.fixture
    def http_client(self, config):
        """Create an HttpClient instance."""
        return HttpClient(config)

    def test_initialization(self, http_client, config):
        """Test HttpClient initialization."""
        assert http_client.config == config
        assert http_client.session is not None
        assert http_client.session.headers["X-API-Key"] == "anc_test_key"
        assert http_client.session.headers["User-Agent"] == "anchorai-python/1.0.0"

    def test_request_success(self, http_client):
        """Test successful HTTP request."""
        mock_response = Mock()
        mock_response.ok = True
        mock_response.status_code = 200
        mock_response.content = b'{"result": "success"}'
        mock_response.json.return_value = {"result": "success"}

        with patch.object(http_client.session, "request", return_value=mock_response):
            result = http_client.request("GET", "/v1/test")
            assert result == {"result": "success"}

    def test_request_empty_response(self, http_client):
        """Test request with empty response."""
        mock_response = Mock()
        mock_response.ok = True
        mock_response.status_code = 200
        mock_response.content = b""

        with patch.object(http_client.session, "request", return_value=mock_response):
            result = http_client.request("GET", "/v1/test")
            assert result == {}

    def test_request_with_data(self, http_client):
        """Test request with JSON data."""
        mock_response = Mock()
        mock_response.ok = True
        mock_response.status_code = 200
        mock_response.content = b'{"id": "123"}'
        mock_response.json.return_value = {"id": "123"}

        with patch.object(http_client.session, "request", return_value=mock_response):
            result = http_client.request("POST", "/v1/test", data={"name": "test"})
            assert result == {"id": "123"}
            # Verify JSON was sent
            http_client.session.request.assert_called_once()
            call_kwargs = http_client.session.request.call_args[1]
            assert call_kwargs["json"] == {"name": "test"}

    def test_request_with_params(self, http_client):
        """Test request with query parameters."""
        mock_response = Mock()
        mock_response.ok = True
        mock_response.status_code = 200
        mock_response.content = b'{"items": []}'
        mock_response.json.return_value = {"items": []}

        with patch.object(http_client.session, "request", return_value=mock_response):
            result = http_client.request("GET", "/v1/test", params={"limit": 10})
            assert result == {"items": []}
            # Verify params were sent
            http_client.session.request.assert_called_once()
            call_kwargs = http_client.session.request.call_args[1]
            assert call_kwargs["params"] == {"limit": 10}

    def test_handle_error_400_validation(self, http_client):
        """Test handling 400 ValidationError."""
        mock_response = Mock()
        mock_response.status_code = 400
        mock_response.json.return_value = {
            "message": "Invalid input",
            "field": "name",
        }
        mock_response.text = "Invalid input"

        with pytest.raises(ValidationError) as exc_info:
            http_client._handle_error(mock_response)
        assert exc_info.value.message == "Invalid input"
        assert exc_info.value.field == "name"

    def test_handle_error_401_authentication(self, http_client):
        """Test handling 401 AuthenticationError."""
        mock_response = Mock()
        mock_response.status_code = 401
        mock_response.json.return_value = {"message": "Invalid API key"}
        mock_response.text = "Invalid API key"

        with pytest.raises(AuthenticationError) as exc_info:
            http_client._handle_error(mock_response)
        assert exc_info.value.message == "Invalid API key"

    def test_handle_error_403_authorization(self, http_client):
        """Test handling 403 AuthorizationError."""
        mock_response = Mock()
        mock_response.status_code = 403
        mock_response.json.return_value = {
            "message": "Permission denied",
            "required_permission": "write:data",
        }
        mock_response.text = "Permission denied"

        with pytest.raises(AuthorizationError) as exc_info:
            http_client._handle_error(mock_response)
        assert exc_info.value.message == "Permission denied"
        assert exc_info.value.required_permission == "write:data"

    def test_handle_error_404_not_found(self, http_client):
        """Test handling 404 NotFoundError."""
        mock_response = Mock()
        mock_response.status_code = 404
        mock_response.json.return_value = {"message": "Resource not found"}
        mock_response.text = "Resource not found"

        with pytest.raises(NotFoundError) as exc_info:
            http_client._handle_error(mock_response)
        assert exc_info.value.message == "Resource not found"

    def test_handle_error_429_rate_limit(self, http_client):
        """Test handling 429 RateLimitError."""
        mock_response = Mock()
        mock_response.status_code = 429
        mock_response.json.return_value = {
            "message": "Rate limit exceeded",
            "retry_after": 60,
        }
        mock_response.text = "Rate limit exceeded"

        with pytest.raises(RateLimitError) as exc_info:
            http_client._handle_error(mock_response)
        assert exc_info.value.message == "Rate limit exceeded"
        assert exc_info.value.retry_after == 60

    def test_handle_error_500_server_error(self, http_client):
        """Test handling 500 ServerError."""
        mock_response = Mock()
        mock_response.status_code = 500
        mock_response.json.return_value = {"message": "Internal server error"}
        mock_response.text = "Internal server error"

        with pytest.raises(ServerError) as exc_info:
            http_client._handle_error(mock_response)
        assert exc_info.value.message == "Internal server error"

    def test_handle_error_policy_violation(self, http_client):
        """Test handling policy violation error."""
        mock_response = Mock()
        mock_response.status_code = 400
        mock_response.json.return_value = {
            "message": "Blocked by policy",
            "blocked_by": "pii_filter",
        }
        mock_response.text = "Blocked by policy"

        with pytest.raises(PolicyViolationError) as exc_info:
            http_client._handle_error(mock_response)
        assert exc_info.value.message == "Blocked by policy"
        assert exc_info.value.policy_name == "pii_filter"

    def test_handle_error_invalid_json(self, http_client):
        """Test handling error with invalid JSON response."""
        mock_response = Mock()
        mock_response.status_code = 500
        mock_response.json.side_effect = ValueError("Invalid JSON")
        mock_response.text = "Server error"

        with pytest.raises(ServerError) as exc_info:
            http_client._handle_error(mock_response)
        assert "Server error" in exc_info.value.message

    def test_retry_on_server_error(self, http_client):
        """Test retry logic on server errors."""
        mock_response_fail = Mock()
        mock_response_fail.ok = False
        mock_response_fail.status_code = 500
        mock_response_fail.json.return_value = {"message": "Server error"}
        mock_response_fail.text = "Server error"

        mock_response_success = Mock()
        mock_response_success.ok = True
        mock_response_success.status_code = 200
        mock_response_success.content = b'{"result": "success"}'
        mock_response_success.json.return_value = {"result": "success"}

        with patch.object(
            http_client.session,
            "request",
            side_effect=[mock_response_fail, mock_response_success],
        ), patch("time.sleep"):  # Don't actually sleep in tests
            result = http_client.request("GET", "/v1/test")
            assert result == {"result": "success"}
            assert http_client.session.request.call_count == 2

    def test_retry_on_timeout(self, http_client):
        """Test retry logic on timeout."""
        mock_response = Mock()
        mock_response.ok = True
        mock_response.status_code = 200
        mock_response.content = b'{"result": "success"}'
        mock_response.json.return_value = {"result": "success"}

        with patch.object(
            http_client.session,
            "request",
            side_effect=[Timeout("Request timeout"), mock_response],
        ), patch("time.sleep"):  # Don't actually sleep in tests
            result = http_client.request("GET", "/v1/test")
            assert result == {"result": "success"}
            assert http_client.session.request.call_count == 2

    def test_retry_on_connection_error(self, http_client):
        """Test retry logic on connection error."""
        mock_response = Mock()
        mock_response.ok = True
        mock_response.status_code = 200
        mock_response.content = b'{"result": "success"}'
        mock_response.json.return_value = {"result": "success"}

        with patch.object(
            http_client.session,
            "request",
            side_effect=[ConnectionError("Connection failed"), mock_response],
        ), patch("time.sleep"):  # Don't actually sleep in tests
            result = http_client.request("GET", "/v1/test")
            assert result == {"result": "success"}
            assert http_client.session.request.call_count == 2

    def test_no_retry_on_client_error(self, http_client):
        """Test that 4xx errors (except 429) don't retry."""
        mock_response = Mock()
        mock_response.ok = False
        mock_response.status_code = 400
        mock_response.json.return_value = {"message": "Bad request"}
        mock_response.text = "Bad request"

        with patch.object(http_client.session, "request", return_value=mock_response):
            with pytest.raises(ValidationError):
                http_client.request("GET", "/v1/test")
            # Should not retry on 400
            assert http_client.session.request.call_count == 1

    def test_retry_on_429_rate_limit(self, http_client):
        """Test that 429 errors do retry."""
        mock_response_fail = Mock()
        mock_response_fail.ok = False
        mock_response_fail.status_code = 429
        mock_response_fail.json.return_value = {
            "message": "Rate limit",
            "retry_after": 1,
        }
        mock_response_fail.text = "Rate limit"

        mock_response_success = Mock()
        mock_response_success.ok = True
        mock_response_success.status_code = 200
        mock_response_success.content = b'{"result": "success"}'
        mock_response_success.json.return_value = {"result": "success"}

        with patch.object(
            http_client.session,
            "request",
            side_effect=[mock_response_fail, mock_response_success],
        ), patch("time.sleep"):  # Don't actually sleep in tests
            result = http_client.request("GET", "/v1/test")
            assert result == {"result": "success"}
            assert http_client.session.request.call_count == 2

    def test_max_retries_exceeded(self, http_client):
        """Test that NetworkError is raised after max retries."""
        with patch.object(
            http_client.session,
            "request",
            side_effect=ConnectionError("Connection failed"),
        ), patch("time.sleep"):  # Don't actually sleep in tests
            with pytest.raises(NetworkError) as exc_info:
                http_client.request("GET", "/v1/test")
            assert "Connection error" in str(exc_info.value)
            # Should retry 3 times (retry_attempts=3) + initial attempt = 4 total
            assert http_client.session.request.call_count == 4

    def test_exponential_backoff(self, http_client):
        """Test that exponential backoff is used."""
        mock_response = Mock()
        mock_response.ok = True
        mock_response.status_code = 200
        mock_response.content = b'{"result": "success"}'
        mock_response.json.return_value = {"result": "success"}

        with patch.object(
            http_client.session,
            "request",
            side_effect=[ConnectionError("Connection failed"), mock_response],
        ), patch("time.sleep") as mock_sleep:
            http_client.request("GET", "/v1/test")
            # Should sleep with exponential backoff: retry_delay * (2^attempt)
            # First retry: 0.1 * (2^0) = 0.1
            mock_sleep.assert_called_once_with(0.1)

    def test_get_method(self, http_client):
        """Test GET convenience method."""
        mock_response = Mock()
        mock_response.ok = True
        mock_response.status_code = 200
        mock_response.content = b'{"result": "success"}'
        mock_response.json.return_value = {"result": "success"}

        with patch.object(http_client.session, "request", return_value=mock_response):
            result = http_client.get("/v1/test", params={"limit": 10})
            assert result == {"result": "success"}
            http_client.session.request.assert_called_once_with(
                "GET",
                "https://api.getanchor.dev/v1/test",
                json=None,
                params={"limit": 10},
                timeout=30.0,
                verify=True,
            )

    def test_post_method(self, http_client):
        """Test POST convenience method."""
        mock_response = Mock()
        mock_response.ok = True
        mock_response.status_code = 200
        mock_response.content = b'{"id": "123"}'
        mock_response.json.return_value = {"id": "123"}

        with patch.object(http_client.session, "request", return_value=mock_response):
            result = http_client.post("/v1/test", data={"name": "test"})
            assert result == {"id": "123"}
            http_client.session.request.assert_called_once_with(
                "POST",
                "https://api.getanchor.dev/v1/test",
                json={"name": "test"},
                params=None,
                timeout=30.0,
                verify=True,
            )

    def test_put_method(self, http_client):
        """Test PUT convenience method."""
        mock_response = Mock()
        mock_response.ok = True
        mock_response.status_code = 200
        mock_response.content = b'{"updated": true}'
        mock_response.json.return_value = {"updated": True}

        with patch.object(http_client.session, "request", return_value=mock_response):
            result = http_client.put("/v1/test", data={"name": "updated"})
            assert result == {"updated": True}

    def test_patch_method(self, http_client):
        """Test PATCH convenience method."""
        mock_response = Mock()
        mock_response.ok = True
        mock_response.status_code = 200
        mock_response.content = b'{"patched": true}'
        mock_response.json.return_value = {"patched": True}

        with patch.object(http_client.session, "request", return_value=mock_response):
            result = http_client.patch("/v1/test", data={"name": "patched"})
            assert result == {"patched": True}

    def test_delete_method(self, http_client):
        """Test DELETE convenience method."""
        mock_response = Mock()
        mock_response.ok = True
        mock_response.status_code = 200
        mock_response.content = b'{"deleted": true}'
        mock_response.json.return_value = {"deleted": True}

        with patch.object(http_client.session, "request", return_value=mock_response):
            result = http_client.delete("/v1/test")
            assert result == {"deleted": True}

